#include "stdafx.h"

#include "gen_typs.h"
#include "gen_util.h"
#include "osd_exp.h"

#if defined( ESTIMATE_CODE_SIZE )
#pragma code_seg( ".eng_code" )
#endif

#define MOVE_UNALIGNED32_TO_32( d, s )	( *(U32 *) (void*) (d) = *(U32 *) (void*) (s) )

#if VDEBUG || !defined(_X8616B)
void
gen_DumpBuffer(
	U8*		Buffer,
	U32		Count,
	U32		Display,
	U32		starting )
{
	U32 i = 0;
	U32 j;
	U32 Temp32;
	U8	BufChar;

	if( (Count < 4) || (Count & 0x01) )
	{
		Display = DB_BYTE_DISPLAY;
	}

	/* OSD_DPRINT( ("\nOffset Value\n") ); ? Nasty little dump buffer routine! */
	while( i < Count )
	{
		/* Print current offset */
		OSD_DPRINT( ("%05X    ", i + starting) );

		/* Print 16 hex chars */
		for( j = 0; j < 16; )
		{
			if( i + j >= Count )
			{
				OSD_DPRINT( ("\n") );
				return;
			}

			/* Make sure that the INT8 doesn't get sign-extended! */
			switch( Display )
			{
			/* Default is BYTE display */
			default:
				OSD_DPRINT( ("%02X ", *((U8*) &Buffer[i + j])) );
				j += 1;
				break;

			case DB_WORD_DISPLAY:
				MOVE_UNALIGNED32_TO_32( &Temp32, &Buffer[i + j] );
				OSD_DPRINT( ("%04X ", (unsigned short) Temp32) );
				j += 2;
				break;

			case DB_DWORD_DISPLAY:
				MOVE_UNALIGNED32_TO_32( &Temp32, &Buffer[i + j] );
				OSD_DPRINT( ("%08X ", Temp32) );
				j += 4;
				break;

			case DB_QWORD_DISPLAY:
				MOVE_UNALIGNED32_TO_32( &Temp32, &Buffer[i + j] );
				OSD_DPRINT( ("%08X", Temp32) );

				MOVE_UNALIGNED32_TO_32( &Temp32, &Buffer[i + j + 4] );
				OSD_DPRINT( ("%08X ", Temp32) );
				j += 8;
				break;
			}
		}

		/*
		 * Print the ASCII equivalent characters But watch out for the bad unprintable
		 * ones...
		 */
		for( j = 0; j < 16; j++ )
		{
			if( i + j >= Count )
			{
				OSD_DPRINT( ("\n") );
				return;
			}

			BufChar = Buffer[i + j];
			if
			(
				(BufChar > 0x1F && BufChar < 0x2E)
			||	(BufChar > 0x2F && BufChar < 0x61)
			||	(BufChar > 0x60 && BufChar < 0x7F)
			)
			{
				OSD_DPRINT( ("%c", BufChar) );
			}
			else
			{
				OSD_DPRINT( (".") );
			}
		}

		/* Done with that line. */
		OSD_DPRINT( ("\n") );
		i += 16;
	}
}
#endif

#if !defined(_X8616B)
/*****************************************************************
 *		General Sorted Single Linked List Implementation
 *****************************************************************/
#define GEN_SORTED_SLIST_ISEMPTY(ListHeader)	(ListHeader->pHeader==NULL)

#define GEN_SORTED_SLIST_KEY(ListHead, ListEntry)				\
	*(U32*)(((PU8)ListEntry)+(ListHead->KeyOffsetInListItem))

#define GEN_SORTED_SLIST_NEXTPOINTER(ListHead, ListEntry)		\
	*((bpVOID*)(((PU8)ListEntry)+(ListHead->NextPointerOffsetInListItem)))

void 
Gen_Sorted_SList_Initialize(
	bpGEN_SORTED_SLIST_HEADER ListHeader,
	U32 NextPointerOffsetInListItem,
	U32 KeyOffsetInListItem
	)
{
	ListHeader->pHeader = NULL;
	ListHeader->NextPointerOffsetInListItem = NextPointerOffsetInListItem;
	ListHeader->KeyOffsetInListItem = KeyOffsetInListItem;
}

void
Gen_Sorted_SList_Insert(
	bpGEN_SORTED_SLIST_HEADER ListHeader,
	bpVOID pEntry
	)
{
	bpVOID pPrev = NULL;
	bpVOID pCurr = NULL;

	if ( GEN_SORTED_SLIST_ISEMPTY(ListHeader) )		/* Empty List */
	{
		ListHeader->pHeader = pEntry;
		return;
	}
	
	/* For the first item, pPrev==NULL */
	pCurr = ListHeader->pHeader;
	if ( GEN_SORTED_SLIST_KEY(ListHeader, pEntry) <=
		GEN_SORTED_SLIST_KEY(ListHeader, pCurr) )
	{	
		/* New List header*/
		GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pEntry)
			= ListHeader->pHeader;
		ListHeader->pHeader = pEntry;
		return;
	}
	else
	{
		pPrev = pCurr;
		pCurr = GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pCurr);	/* Can be NULL */
	}

	while ( pCurr!=NULL )
	{
		if ( GEN_SORTED_SLIST_KEY(ListHeader, pEntry) <=
			GEN_SORTED_SLIST_KEY(ListHeader, pCurr) )
		{
			/* It's the place to insert. */
			GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pPrev)
				= pEntry;
			GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pEntry)
				= pCurr;
			return;
		}
		else
		{
			pPrev = pCurr;
			pCurr = GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pCurr);
		}
	}

	/* The New tail */
	GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pPrev)	= pEntry;
	return;
}

/* Just return the list header, it's not removed. */
bpVOID
Gen_Sorted_SList_GetFirst(
	bpGEN_SORTED_SLIST_HEADER ListHeader
	)
{
	return ListHeader->pHeader;
}

/* Just return the next entry, it's not removed. */
bpVOID 
Gen_Sorted_SList_GetNext(
	bpGEN_SORTED_SLIST_HEADER ListHeader,
	bpVOID pEntry
	)
{
	return GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pEntry);
}

BoOL
Gen_Sorted_SList_IsEmpty(
	bpGEN_SORTED_SLIST_HEADER ListHeader
	)
{
	return GEN_SORTED_SLIST_ISEMPTY(ListHeader);
}

bpVOID
Gen_Sorted_SList_RemoveHead(
	bpGEN_SORTED_SLIST_HEADER ListHeader
	)
{
	bpVOID pEntry = ListHeader->pHeader;
	if ( pEntry==NULL )
		return NULL;

	ListHeader->pHeader = 
		GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, ListHeader->pHeader);

	GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pEntry) = NULL;
	
	return pEntry;
}
BoOL
Gen_Sorted_SList_CheckExist(
	bpGEN_SORTED_SLIST_HEADER ListHeader,
	bpVOID pEntry
	)
{
	bpVOID pCurr = ListHeader->pHeader;
	while ( pCurr != NULL )
	{
		if ( pCurr == pEntry )
			return TRUE;
		else
			pCurr = GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pCurr);
	}

	return FALSE;
}

bpVOID
Gen_Sorted_SList_RemoveEntry(
	bpGEN_SORTED_SLIST_HEADER ListHeader,
	bpVOID pEntry
	)
{
	bpVOID pPrev = NULL;
	bpVOID pCurr = NULL;
	
	pCurr = ListHeader->pHeader;

	while ( pCurr != NULL )
	{
		if ( pCurr == pEntry )	/* Remove this entry. */
		{
			if ( pPrev==NULL )	/* pCurr is the first Entry */
			{	
				Gen_Sorted_SList_RemoveHead(ListHeader);
				return pCurr;
			}
			else
			{
				GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pPrev)
					= GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pCurr);

				GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pCurr) = NULL;
				return pCurr;
			}
		} 
		else
		{
			pPrev = pCurr;
			pCurr = GEN_SORTED_SLIST_NEXTPOINTER(ListHeader, pCurr);
		}
	}

	return NULL;
}

void 
Gen_Sorted_SList_Clear(
	bpGEN_SORTED_SLIST_HEADER ListHeader
	)
{
	/* Clear the bpGEN_SORTED_SLIST_HEADER data structure. */
	ListHeader->pHeader = NULL;
	ListHeader->NextPointerOffsetInListItem = 0;
	ListHeader->KeyOffsetInListItem = 0;

	/* Somewhere the Entry should be recycled. */
}
#endif

/*
 * TODO: following functions need assembly optimization
 */
void gen_bzero(
	void* Buffer, 
	U32 ByteCount )
{
	U32* pInt = (U32*) Buffer;
	U32 i;

	for( i=0; i<ByteCount/sizeof(U32); i++ )
	{
		*pInt++ = 0L;
	}

	if( ByteCount%sizeof(U32) )
	{
		U8* pU8 = (U8*) pInt;
		for( i=0; i<ByteCount%sizeof(U32); i++ )
			*pU8++ = 0;
	}
}

void gen_memmove(void* pDest, const void* pSrc, U32 ByteCount)
{
	U32 i;
	U8* pU8Src;
	U8* pU8Dst;

	if( ((PTR_SIZE)pDest)>=((PTR_SIZE)pSrc)
		&& ((PTR_SIZE)pDest)<((PTR_SIZE)pSrc)+ByteCount )
	{
		/* overalap... */
byte_by_byte:
		pU8Src = ((U8*) pSrc) + ByteCount - 1;
		pU8Dst = ((U8*) pDest) + ByteCount - 1;

		for( i=0; i<ByteCount; i++ )
			*pU8Dst-- = *pU8Src--;
	}
	else
	{
		if( (((PTR_SIZE)pSrc)&0x03) || (((PTR_SIZE)pDest)&0x03) )
			goto byte_by_byte;
		else
		{
			U32* pIntSrc = (U32*) pSrc;
			U32* pIntDst = (U32*) pDest;

			for( i=0; i<ByteCount/sizeof(U32); i++ )
			{
				*pIntDst++ = *pIntSrc++;
			}

			if( ByteCount%sizeof(U32) )
			{
				pU8Src = (U8*) pIntSrc;
				pU8Dst = (U8*) pIntDst;
				for( i=0; i<ByteCount%sizeof(U32); i++ )
					*pU8Dst++ = *pU8Src++;
			}
		}
	}
}
