/***************************************************************************
 *
 * Module		: Component Management ( CMPM )
 *
 * Description	: Gather resource requirement and allocate resource.
 *	Start stop each components.
 *	Currently we only have two components, the OSD and the Message Unit.
 *	Later in order to tune performance, we may expect Cache module here.
 *
 * To Do List	:
 *
 ***************************************************************************/
#include "stdafx.h"
#include "fmu_scsi.h"

bpOSD_COMP_EXTENSION gOSDExtension = NULL;

struct _ComponentInterface IComp[HOST_COMP_TYPE_MAXIMUM] = {
	{ /* comp_osd */
		"OSD",
		OSD_ComponentGetExtensionSize, 
		OSD_ComponentGetDMASafeMemoryRequirement, 
		OSD_ComponentInitialize, 
		OSD_ComponentStart,
		OSD_ComponentStop,
		OSD_ComponentRestart,
	},
	{ /* comp_mu */
		"MU",
		MU_ComponentGetExtensionSize, 
		MU_ComponentGetDMASafeMemoryRequirement, 
		MU_ComponentInitialize, 
		MU_ComponentStart,
		MU_ComponentStop,
		MU_ComponentRestart
	}
};

U32
CMPM_GetComponentExtensionRequirement(
	BoOL isHibernation
	)
{
	MEMORY_REQUIREMENT totalRequirement = { 0, 0 };
	MEMORY_REQUIREMENT compRequirement;
	U16 CompIndex; 

	for( CompIndex=0; CompIndex<HOST_COMP_TYPE_MAXIMUM; CompIndex++ )
	{
		if( IComp[CompIndex].COMP_GET_EXT_REQUIREMENT )
		{
			IComp[CompIndex].COMP_GET_EXT_REQUIREMENT(&compRequirement);
			totalRequirement.MinRequirement += compRequirement.MinRequirement;
			totalRequirement.MaxRequirement += compRequirement.MaxRequirement;
		}
	}
	totalRequirement.MinRequirement += HOST_COMP_TYPE_MAXIMUM * COMPONENT_HEADER_SIZE;
	totalRequirement.MaxRequirement += HOST_COMP_TYPE_MAXIMUM * COMPONENT_HEADER_SIZE;

	OSD_DPRINT( ("Required extension size: max: %d Min: %d\n", 
		totalRequirement.MaxRequirement, totalRequirement.MinRequirement) );

	return (isHibernation 
		? totalRequirement.MinRequirement
		: totalRequirement.MaxRequirement);
}

BoOL
CMPM_AssignComponentExtension(
	IN bpOSD_COMP_EXTENSION pOSDExtension,
	IN LPVOID DeviceExtension,
	IN U32 ExtensionSize
	)
{
	/* Check the DeviceExtension Size */
	U32 extRequirement = 
		CMPM_GetComponentExtensionRequirement(gIsHibernation);

	PTR_SIZE pTemp = ((PTR_SIZE)DeviceExtension + COMPONENT_HEADER_SIZE);
	MEMORY_REQUIREMENT compRequirement;
	U16 CompIndex;
	PCOMPONENT pCompHeader = NULL;
	U32 assignSize = 0;

	OSD_DASSERT(ExtensionSize>=extRequirement);

	/* Assign Component Extension */
	for( CompIndex=0; CompIndex<HOST_COMP_TYPE_MAXIMUM; CompIndex++ )
	{
		if( IComp[CompIndex].COMP_GET_EXT_REQUIREMENT )
		{
			IComp[CompIndex].COMP_GET_EXT_REQUIREMENT(&compRequirement);
			pOSDExtension->ComponentResource[CompIndex].ComponentExtension = (LPVOID)pTemp;
			OSD_TRACE(("CMPM_AssignComponentExtension:%x %x\n",CompIndex,pOSDExtension->ComponentResource[CompIndex].ComponentExtension));

			/* Assign the back pointer to OSD Extension as well. */
			pCompHeader = (PCOMPONENT)(pTemp-COMPONENT_HEADER_SIZE);
			pCompHeader->OSDExtension = pOSDExtension;
			pCompHeader->ComponentType = CompIndex;
			pCompHeader->HeaderSize = COMPONENT_HEADER_SIZE;

			if ( gIsHibernation )
				assignSize = compRequirement.MinRequirement;
			else
				assignSize = compRequirement.MaxRequirement;
			
			pTemp += COMPONENT_HEADER_SIZE + assignSize;
			pOSDExtension->ComponentResource[CompIndex].ComponentExtensionSize = assignSize;
			pCompHeader->ExtensionSize = assignSize;
		}
	}

	return TRUE;
}

U32
CMPM_GetUncachedMemoryRequirement(
	IN bpOSD_COMP_EXTENSION pOSDExtension,
	IN BoOL isHibernation
	)
{
	U32 DeviceID = pOSDExtension->ControllerInfo.DeviceID;
	U32 SubDeviceID = pOSDExtension->ControllerInfo.SubDeviceID;
	U16 CompIndex;

	MEMORY_REQUIREMENT totalRequirement = { 0, 0 };
	MEMORY_REQUIREMENT compRequirement;

	for( CompIndex=0; CompIndex<HOST_COMP_TYPE_MAXIMUM; CompIndex++ )
	{
		if( IComp[CompIndex].COMP_GET_DMA_REQUIREMENT )
		{
			IComp[CompIndex].COMP_GET_DMA_REQUIREMENT(
				DeviceID, SubDeviceID, &compRequirement);
			totalRequirement.MinRequirement += compRequirement.MinRequirement;
			totalRequirement.MaxRequirement += compRequirement.MaxRequirement;
		}
	}

	OSD_DPRINT( ("Required DMA safe size: max: %d Min: %d\n", 
		totalRequirement.MaxRequirement, totalRequirement.MinRequirement) );

	return (isHibernation 
		? totalRequirement.MinRequirement
		: totalRequirement.MaxRequirement);
}

BoOL
CMPM_AssignUncachedMemory(
	IN bpOSD_COMP_EXTENSION pOSDExtension,
	IN LPVOID VirtualAddress,
	IN PHYADDR PhysicalAddress,
	IN U32 MemorySize
	)
{
	/* Check the Uncached Memory Size */
	U32 DeviceID = pOSDExtension->ControllerInfo.DeviceID;
	U32 SubDeviceID = pOSDExtension->ControllerInfo.SubDeviceID;

	U32 requirement = 
		CMPM_GetUncachedMemoryRequirement(pOSDExtension, gIsHibernation);

	PTR_SIZE pTempV = (PTR_SIZE)VirtualAddress;
	PHYADDR pTempP = PhysicalAddress;
	MEMORY_REQUIREMENT compRequirement;
	U16 CompIndex;

	OSD_DASSERT(MemorySize>=requirement);
	OSD_TRACE(("CMPM_AssignUncachedMemory:%x %x\n",1,pOSDExtension->ComponentResource[1].ComponentExtension));

	/* Assign Uncached Memory */
	for( CompIndex=0; CompIndex<HOST_COMP_TYPE_MAXIMUM; CompIndex++ )
	{
		if( IComp[CompIndex].COMP_GET_DMA_REQUIREMENT )
		{
			IComp[CompIndex].COMP_GET_DMA_REQUIREMENT(DeviceID, SubDeviceID, &compRequirement);
			pOSDExtension->ComponentResource[CompIndex].DMASafeAddress = (LPVOID)pTempV;
			pOSDExtension->ComponentResource[CompIndex].DMASafePhysicalAddress = pTempP;

			if ( gIsHibernation )
			{
				pOSDExtension->ComponentResource[CompIndex].DMASafeSize =
					compRequirement.MinRequirement;
				pTempV += compRequirement.MinRequirement;
				pTempP.u.Low += compRequirement.MinRequirement;
			}
			else
			{
				pOSDExtension->ComponentResource[CompIndex].DMASafeSize =
					compRequirement.MaxRequirement;
				pTempV += compRequirement.MaxRequirement;
				pTempP.u.Low += compRequirement.MaxRequirement;
			}
		}
	}

	OSD_TRACE(("CMPM_AssignUncachedMemory:%x %x\n",1,pOSDExtension->ComponentResource[1].ComponentExtension));
	return TRUE;
}

/* It includes DIMM and DMA Safe memory. 
 * DeviceExtension has been assigned already at the very beginning. 
 */
BoOL
CMPM_AssignComponentResource(
	IN LPVOID  DeviceExtension
	)
{
	bpOSD_COMP_EXTENSION pOSDExtension = 
		CMPM_GetOSDExtensionFromDeviceExtension(DeviceExtension);
	BoOL result;

	/* Uncached Memory */
	LPVOID VirtualAddress = pOSDExtension->DMASafeAddress;
	PHYADDR PhysicalAddress = pOSDExtension->DMASafePhysicalAddress;
	U32 MemorySize = pOSDExtension->DMASafeSize;	

	/* Assign Uncached Memory */
	result = CMPM_AssignUncachedMemory(
				pOSDExtension,
				VirtualAddress,
				PhysicalAddress,
				MemorySize
			);
	OSD_DASSERT(result);

	return TRUE;
}

/* Initialize the Components. */
BoOL 
CMPM_InitializeComponents(
	IN OUT bpOSD_COMP_EXTENSION pOSDExtension,
	IN OUT LPVOID ConfigInfo,	/* PPORT_CONFIGURATION_INFORMATION */
	IN U32 DeviceID,
	IN U32 SubDeviceID
	)
{
	MEMORY_AVAILABLE memAvailable;
	PCOMP_RESOURCE pCompResource;
	U16 CompIndex;
	U8 i;
	LPVOID pContext = NULL;
	MU_INITIALIZE_CONTEXT muInitContext;

	/* OSD Component is the one which is served first. 
	 * The order ( the index ) of each Compnent is very important.
	 */

	for( CompIndex=0; CompIndex<HOST_COMP_TYPE_MAXIMUM; CompIndex++ )
	{
		if( IComp[CompIndex].COMP_INITIALIZE )
		{
			pCompResource = &pOSDExtension->ComponentResource[CompIndex];
			memAvailable.ComponentExtension = pCompResource->ComponentExtension ;
			memAvailable.ComponentExtensionSize = pCompResource->ComponentExtensionSize ;
			memAvailable.DMASafeAddress = pCompResource->DMASafeAddress ;
			memAvailable.DMASafeSize = pCompResource->DMASafeSize ;
			memAvailable.DMASafePhysicalAddress = pCompResource->DMASafePhysicalAddress ;

			switch ( CompIndex ){
				case COMP_TYPE_OSD:
					pContext = ConfigInfo;
					break;
				case COMP_TYPE_MU:
					for ( i=0; i<BASE_ADDRESS_MAX_NUM; i++ )
						muInitContext.BaseAddress[i] = pOSDExtension->BaseAddress[i];

					pContext = &muInitContext;
					break;
				default:
					pContext = NULL;
					break;
			}

			OSD_TRACE(("CMPM_InitializeComponents:%x %x\n",1,pOSDExtension->ComponentResource[1].ComponentExtension));
			IComp[CompIndex].COMP_INITIALIZE(
				pCompResource->ComponentExtension, 
				&memAvailable, 
				DeviceID, 
				SubDeviceID,
				pContext
				);
			OSD_TRACE(("CMPM_InitializeComponents:%x %x\n",1,pOSDExtension->ComponentResource[1].ComponentExtension));
		}
	}

	return TRUE;
}

void
CMPM_StartComponents(
	bpOSD_COMP_EXTENSION pOSDExtension,
	U16 BeginComp
	)
{
	U16 CompIndex;
	BoOL isRestart = OSDI_IsRestart(pOSDExtension);

	pOSDExtension->State = SYS_STATE_STARTING;

	/* 
		Important: The order of the Component Index must be the same of the order of Start Sequence 
		So: Index: OSD->MU
	*/
	for( CompIndex=BeginComp; CompIndex<HOST_COMP_TYPE_MAXIMUM; CompIndex++ )
	{
		if( IComp[CompIndex].COMP_START )
		{
			if ( pOSDExtension->CompStartStatus&(1<<CompIndex) )
				continue;
			else
			{
				if ( !isRestart)
				{
					IComp[CompIndex].COMP_START(
						pOSDExtension->ComponentResource[CompIndex].ComponentExtension	/* Extension */	
						);
					return;
				}
				else
				{
					IComp[CompIndex].COMP_RESTART(
						pOSDExtension->ComponentResource[CompIndex].ComponentExtension	/* Extension */	
						);
					return;
				}
			}
		}
		else
		{
			pOSDExtension->CompStartStatus |= (1<<CompIndex);
		}
	}
	return;

}

void
CMPM_RestartComponents(
	bpOSD_COMP_EXTENSION pOSDExtension
	)
{
	OSDI_SetRestartFlag(pOSDExtension);
	OSDI_ClearComponentsReadyFlag(pOSDExtension);

	CMPM_StartComponents(pOSDExtension, COMP_TYPE_OSD);
}

VOID 
CMPM_StopComponents(
	bpOSD_COMP_EXTENSION pOSDExtension
	)	
{
	U16 CompIndex;

	/*	Stop all the other Components. Begin at the PBM, end with CAM. 
	  COMP_TYPE_MU==1 COMP_TYPE_OSD==0 */
	for( CompIndex=HOST_COMP_TYPE_MAXIMUM-1; CompIndex>0; CompIndex-- )
	{
		if( IComp[CompIndex].COMP_STOP )
		{
			IComp[CompIndex].COMP_STOP(
				CMPM_GetComponentExtension(pOSDExtension, CompIndex)
				);

			OSD_TRACE(("Component %s is stopped.\n", IComp[CompIndex].CompName));
		}
	}

	/* Stop OSD Component First. Suppose COMP_STOP functions are synchronized. */
	OSD_ComponentStop(pOSDExtension);
    OSD_TRACE(("OSD Component is stopped.\n"));
}

bpVOID 
CMPM_GetComponentExtension(
	LPVOID CallerExtension,
	U16 ComponentType
	)	
{
	bpOSD_COMP_EXTENSION pOSDExtension
		= (bpOSD_COMP_EXTENSION)OSDI_GetOSDExtension(CallerExtension);

	OSD_DASSERT(ComponentType<HOST_COMP_TYPE_MAXIMUM);
	return pOSDExtension->ComponentResource[ComponentType].ComponentExtension;
}

/* Whenever there is a component started, it'll call the OSD_ComponentStarted function. */
VOID
OSD_ComponentStarted(
	LPVOID SenderComponentExtension,
	U32 StartStatus
	)
{	
	U16 ComponentType = 
		OSDI_GetComponentType(SenderComponentExtension);

	bpOSD_COMP_EXTENSION pOSDExtension = (bpOSD_COMP_EXTENSION)
		CMPM_GetComponentExtension(SenderComponentExtension, COMP_TYPE_OSD);

	OSD_DASSERT(StartStatus==COMP_START_STATUS_SUCCESS);

	OSD_TRACE(("Component %s is started.\n", IComp[ComponentType].CompName));

	/* Mark the corresponding component is started. */
	pOSDExtension->CompStartStatus |= (1<<ComponentType);

	if ( !OSDI_ComponentsAreReady(pOSDExtension) )
	{
		/* Start other Component. */
		CMPM_StartComponents(pOSDExtension, ComponentType+1);
	}
	else
	{
		pOSDExtension->State = SYS_STATE_STARTED;
#if defined(_WINDRIVER)
		/* All the Components are started. Is there anything we need do? */
		if ( Host_OSD_IsLdrDump(NULL) )
		{
			/*Check the pending srb*/
			if ( pOSDExtension->pPendingSrb )
			{
				RaidStartIo(pOSDExtension->DeviceExtension, pOSDExtension->pPendingSrb);
			}
		}
		else
		{
		}
#endif
	}
}

/***************************************************************************
 *
 *			Component Requirement for OSD Component Itself.
 *
 ***************************************************************************/
void OSD_ComponentGetExtensionSize(
	OUT bpMEMORY_REQUIREMENT pMemRequirement 
	)
{
	pMemRequirement->MaxRequirement = sizeof(OSD_COMP_EXTENSION);
	pMemRequirement->MinRequirement = sizeof(OSD_COMP_EXTENSION);
}

void
OSD_ComponentGetDMASafeMemoryRequirement(
	IN U32 DeviceID,
	IN U32 SubDeviceID,
	OUT bpMEMORY_REQUIREMENT pMemRequirement 
	)
{
#if defined(_WINDRIVER)
	pMemRequirement->MaxRequirement = 0;
	pMemRequirement->MinRequirement = 0;
#elif defined(_LINUXDRIVER)
	pMemRequirement->MaxRequirement = ( MAX_COPY_SIZE) * (MAX_DATA_BUFFER_COUNT);
	pMemRequirement->MinRequirement = (MAX_COPY_SIZE) * (MAX_DATA_BUFFER_COUNT);
#endif
}

#if defined(_WINDRIVER)
BoOL
OSDI_ProcessBaseAddress(
	IN LPVOID pDeviceExtension,
	IN bpOSD_COMP_EXTENSION pOSDExtension,
	IN OUT PPORT_CONFIGURATION_INFORMATION ConfigInfo
	)
{
	U32 i;
	U8 IoSpaceBitmap;

	/* Based on DeviceID and SubDeviceID, determine the Base Address is in IO space or memory space. */
	IoSpaceBitmap = 0x0;

	for( i=0; i<ConfigInfo->NumberOfAccessRanges; i++ )
	{
		if( !ScsiPortValidateRange(
			(PVOID) pDeviceExtension,
			ConfigInfo->AdapterInterfaceType,
			ConfigInfo->SystemIoBusNumber,
			(*ConfigInfo->AccessRanges)[i].RangeStart,
			(*ConfigInfo->AccessRanges)[i].RangeLength,
			((IoSpaceBitmap & (0x01<<i)) ? TRUE : FALSE) ))
		{
			OSD_DASSERT((0));
		}

		(*ConfigInfo->AccessRanges)[i].RangeInMemory = ((IoSpaceBitmap & (0x01<<i)) ? FALSE : TRUE);

#ifdef CONSTRAIN_SDRAM_BASE_ADDRESS
		if( i==2 )
			(*ConfigInfo->AccessRanges)[i].RangeLength = 0x400000L; /* 4M */
#endif
        
		pOSDExtension->BaseAddress[i] = ScsiPortGetDeviceBase(
			(PVOID) pDeviceExtension,
			ConfigInfo->AdapterInterfaceType,
			ConfigInfo->SystemIoBusNumber,
			(*ConfigInfo->AccessRanges)[i].RangeStart,
			(*ConfigInfo->AccessRanges)[i].RangeLength,
			((*ConfigInfo->AccessRanges)[i].RangeInMemory==FALSE)
			);

		OSD_DPRINT(( "ConfigInfo[%d]=%p.\n", i, (*ConfigInfo->AccessRanges)[i].RangeStart ));
		OSD_DPRINT(( "BaseAddress[%d]=0x%x.\n", i, pOSDExtension->BaseAddress[i] ));
		//OSD_ASSERT( pOSDExtension->BaseAddress[i]!=NULL );
	}

	return TRUE;
}

void
OSD_ComponentInitialize(
	IN bpVOID OsdExt,
	IN bpMEMORY_AVAILABLE pMemAvail,
	IN U32 DeviceID,
	IN U32 SubDeviceID,
	IN bpVOID ComponentSpecificContext
)
{
	bpOSD_COMP_EXTENSION pOSDExtension = (bpOSD_COMP_EXTENSION) OsdExt;
	LPVOID pDeviceExtension = pOSDExtension->DeviceExtension;
	BoOL result;

	PPORT_CONFIGURATION_INFORMATION ConfigInfo = 
		(PPORT_CONFIGURATION_INFORMATION) ComponentSpecificContext;

	gOSDExtension = pOSDExtension;

	/* Step 1: Set controller related variables. */
	pOSDExtension->ControllerInfo.SystemIoBusNumber = ConfigInfo->SystemIoBusNumber;
	pOSDExtension->ControllerInfo.SlotNumber = ConfigInfo->SlotNumber;
	pOSDExtension->ControllerInfo.InterruptLevel = ConfigInfo->BusInterruptLevel;
	pOSDExtension->ControllerInfo.InterruptVector = ConfigInfo->BusInterruptVector;
	pOSDExtension->ControllerInfo.DeviceID = DeviceID;
	pOSDExtension->ControllerInfo.SubDeviceID = SubDeviceID;
	pOSDExtension->CompStartStatus = 0;

	/* Step 2: Get Base Address. */
	result = OSDI_ProcessBaseAddress(pDeviceExtension,pOSDExtension,ConfigInfo);

	OSD_DASSERT(result==TRUE);
}
#elif defined(_LINUXDRIVER)
void
OSD_ComponentInitialize(
	IN bpVOID OsdExt,
	IN bpMEMORY_AVAILABLE pMemAvail,
	IN U32 DeviceID,
	IN U32 SubDeviceID,
	IN bpVOID ComponentSpecificContext
)
{
	bpOSD_COMP_EXTENSION pOSDExtension = (bpOSD_COMP_EXTENSION) OsdExt;
			
	spin_lock_init(&(pOSDExtension->timer_lock) );
	spin_lock_init(&(pOSDExtension->datapool_lock) );
	init_req_pool(pOSDExtension);
	
	init_data_pool( (bpVOID)pOSDExtension, pMemAvail);

	pOSDExtension->CompStartStatus = 0;
	
}
#endif

void
OSD_ComponentStop(
	IN bpVOID OSDExt
	)
{
	bpOSD_COMP_EXTENSION pOSDExtension = (bpOSD_COMP_EXTENSION)OSDExt;
#if defined(_WINDRIVER)
#ifndef USE_STORPORT
	OSD_DASSERT(pOSDExtension->TaskCount == 0);
#endif
#endif
	/* Reset the OSD Data Structure. */
	OSDI_ClearComponentsReadyFlag(pOSDExtension);
	OSDI_ClearRestartFlag(pOSDExtension);
}

void
OSD_ComponentStart(
	IN bpVOID ComponentExtension
	)
{
	OSD_ComponentStarted(ComponentExtension, COMP_START_STATUS_SUCCESS);
}

void 
OSD_ComponentRestart(
	IN bpVOID ComponentExtension
	)
{
	//bpOSD_COMP_EXTENSION pOSDExtension = (bpOSD_COMP_EXTENSION) ComponentExtension;

	/* Nothing else needs done. */
	OSD_ComponentStarted(ComponentExtension, COMP_START_STATUS_SUCCESS);
}
