#ifndef OSD_COMPONENT_MANAGEMENT_H
#define OSD_COMPONENT_MANAGEMENT_H

/* Definition for OSD_ComponentStarted */
#define COMP_START_STATUS_SUCCESS 1
#define COMP_START_STATUS_FAILURE 0

/*
 * Definition for OSD_ComponentState
 */
#define SYS_STATE_INITIALIZING	0x00
#define SYS_STATE_STARTING		0x01
#define SYS_STATE_GOING_SLEEP	0x02
#define SYS_STATE_STARTED		0x03
#define SYS_STATE_FULLY_STARTED	0x04
#define SYS_STATE_SHUTINGDOWN	0x05
#define SYS_STATE_SHUTDOWN		0x06

bpVOID 
CMPM_GetComponentExtension(
	LPVOID CallerExtension,
	U16 ComponentType
	);

U32
CMPM_GetComponentExtensionRequirement(
	BoOL isHibernation
	);
BoOL
CMPM_AssignComponentExtension(
	IN bpOSD_COMP_EXTENSION pOSDExtension,
	IN PVOID DeviceExtension,
	IN U32 ExtensionSize
	);
U32
CMPM_GetUncachedMemoryRequirement(
	IN bpOSD_COMP_EXTENSION pOSDExtension,
	IN BoOL isHibernation
	);
BoOL
CMPM_AssignUncachedMemory(
	IN bpOSD_COMP_EXTENSION pOSDExtension,
	IN PVOID VirtualAddress,
	IN PHYADDR PhysicalAddress,
	IN U32 MemorySize
	);
BoOL
CMPM_AssignComponentResource(
	IN PVOID  DeviceExtension
	);
BoOL 
CMPM_InitializeComponents(
	IN OUT bpOSD_COMP_EXTENSION pOSDExtension,
	IN OUT LPVOID ConfigInfo,	/* PPORT_CONFIGURATION_INFORMATION */
	IN U32 DeviceID,
	IN U32 SubDeviceID
	);
void
CMPM_StartComponents(
	bpOSD_COMP_EXTENSION pOSDExtension,
	U16 BeginComp
	);
void
CMPM_RestartComponents(
	bpOSD_COMP_EXTENSION pOSDExtension
	);
VOID 
CMPM_StopComponents(
	bpOSD_COMP_EXTENSION pOSDExtension
	);

void 
OSD_ComponentGetExtensionSize(
	OUT bpMEMORY_REQUIREMENT pMemRequirement 
	);

void
OSD_ComponentGetDMASafeMemoryRequirement(
	IN U32 DeviceID,
	IN U32 SubDeviceID,
	OUT bpMEMORY_REQUIREMENT pMemRequirement 
	);

void
OSD_ComponentInitialize(
	IN bpVOID OsdExt,
	IN bpMEMORY_AVAILABLE pMemAvail,
	IN U32 DeviceID,
	IN U32 SubDeviceID,
	IN bpVOID ComponentSpecificContext
	);

void
OSD_ComponentStop(
	IN bpVOID OSDExt
	);

void
OSD_ComponentStart(
	IN bpVOID ComponentExtension
	);

void 
OSD_ComponentRestart(
	IN bpVOID ComponentExtension
	);

VOID
OSD_ComponentStarted(
	LPVOID SenderComponentExtension,
	U32 StartStatus
	);

struct _ComponentInterface
{
	const char* CompName;	/* component name, debugging convinience */
	void (*COMP_GET_EXT_REQUIREMENT)(bpMEMORY_REQUIREMENT);
	void (*COMP_GET_DMA_REQUIREMENT)(U32, U32, bpMEMORY_REQUIREMENT);
	void (*COMP_INITIALIZE)(PVOID,bpMEMORY_AVAILABLE,U32,U32,PVOID);
	void (*COMP_START)(bpVOID);
	void (*COMP_STOP)(bpVOID);
	void (*COMP_RESTART)(bpVOID);
};

/* bit(0):OSD, bit(1):MU */
#define COMPS_ALL_STARTED	(bit(0)|bit(1))
/* If bit(7) is set, it's during restart procedure. */
#define COMPS_RESTART bit(7)	

#define OSDI_ComponentsAreReady(pOSDExtension) ( ((pOSDExtension->CompStartStatus&(~COMPS_RESTART))&COMPS_ALL_STARTED)==COMPS_ALL_STARTED )
#define OSDI_ClearComponentsReadyFlag(pOSDExtension) pOSDExtension->CompStartStatus&=~COMPS_ALL_STARTED;

#define OSDI_IsRestart(pOSDExtension) ( (pOSDExtension->CompStartStatus&COMPS_RESTART)==COMPS_RESTART )
#define OSDI_SetRestartFlag(pOSDExtension) pOSDExtension->CompStartStatus|=COMPS_RESTART;
#define OSDI_ClearRestartFlag(pOSDExtension) pOSDExtension->CompStartStatus&=~COMPS_RESTART;

#define COMPONENT_EXTENSION_OFFSET													\
	FIELD_OFFSET(COMPONENT, ComponentExtension)

#define	COMPONENT_HEADER_SIZE	COMPONENT_EXTENSION_OFFSET

/* Get the Component Header for each Component Extension */
#define OSDI_GetComponentHeader(pComponentExtension)								\
	(LPVOID)((PTR_SIZE)(pComponentExtension)- COMPONENT_EXTENSION_OFFSET)

/* Return the Component Type */
#define	OSDI_GetComponentType(ComponentExtension)									\
	((PCOMPONENT)(OSDI_GetComponentHeader(ComponentExtension)))->ComponentType

#define CMPM_GetOSDExtensionFromDeviceExtension(pDeviceExtension)					\
	(bpOSD_COMP_EXTENSION)((PTR_SIZE)pDeviceExtension + COMPONENT_EXTENSION_OFFSET)

/* From any kind of Component Extension, get the OSD Extension */
#define OSDI_GetOSDExtension(pComponentExtension)									\
	((PCOMPONENT)(OSDI_GetComponentHeader(pComponentExtension)))->OSDExtension

#endif
