/*
   SuperTrak EX8350/8300 Storage Controller device driver for Linux.

   Written By: Ed Lin <edl@promisechina.com>

   Copyright (C) 2005 Promise Technology Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   NO WARRANTY
   THE PROGRAM IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OR
   CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED INCLUDING, WITHOUT
   LIMITATION, ANY WARRANTIES OR CONDITIONS OF TITLE, NON-INFRINGEMENT,
   MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. Each Recipient is
   solely responsible for determining the appropriateness of using and
   distributing the Program and assumes all risks associated with its
   exercise of rights under this Agreement, including but not limited to
   the risks and costs of program errors, damage to or loss of data,
   programs or equipment, and unavailability or interruption of operations.

   DISCLAIMER OF LIABILITY
   NEITHER RECIPIENT NOR ANY CONTRIBUTORS SHALL HAVE ANY LIABILITY FOR ANY
   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   DAMAGES (INCLUDING WITHOUT LIMITATION LOST PROFITS), HOWEVER CAUSED AND
   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
   TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
   USE OR DISTRIBUTION OF THE PROGRAM OR THE EXERCISE OF ANY RIGHTS GRANTED
   HEREUNDER, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "napa_cfg.h"

static int	gNumAdapter=0;
unsigned char * gShasta_hostp[MAX_ADAPTORS]={0};
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static int	gReleaseAdapter=0;
static unsigned int   gVenID[MAX_ADAPTER_TYPE] = {PCI_VENDOR_ID_INTEL,PCI_VENDOR_ID_INTEL,PCI_VENDOR_ID_PROMISE};
static unsigned int   gDevID[MAX_ADAPTER_TYPE] = {PCI_DEVICE_ID_SHASTA,PCI_DEVICE_ID_SHASTA1,PCI_DEVICE_ID_SHASTA_P};
#endif
static int gShastaMajor;
static int gHighmem_io;
static int gSt_max_lun = 0;
struct proc_dir_entry proc_scsi_shasta = {		
   0,					
   8,					
   "shasta",				
   S_IFDIR | S_IRUGO | S_IXUGO,	
   2,				
};

BoOL gIsHibernation;

extern NEEDCOMPLETE
scsiFlashCommand(
	IN PVOID OSDExt,
	IN PSCSI_REQUEST_BLOCK Srb
);

extern NEEDCOMPLETE
scsiGetControllerInfor(
	IN PVOID OSDExt,
	IN PSCSI_REQUEST_BLOCK Srb
);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static void shasta_intr(int irq, void *dev_id, void *regs);
#else
static irqreturn_t shasta_intr(int irq, void *dev_id, void *regs);
#endif
static PDATA_POOL alloc_data_pool(bpVOID OSDExtension);
static void 
release_data_pool(bpVOID OSDExtension, PDATA_POOL pDataPool);
static PREQ_HEAD get_req_from_pool(bpOSD_COMP_EXTENSION p);
static void put_req_to_pool(bpOSD_COMP_EXTENSION p, PREQ_HEAD pReq);

static int st_open(struct inode *inodep, struct file *filep);
static int st_release(struct inode *inodep, struct file *filep);
static int st_ioctl(struct inode *inodep,struct file *filep,unsigned int req,unsigned long arg);

static int st_fasync(int fd, struct file * filep, int mode);
static char EndRequest(void *cmnd,void *fun,int result);
static void st_set_cmdresult(bpOSD_COMP_EXTENSION p,PSCSI_REQUEST_BLOCK Srb);
static void shasta_req_timer(unsigned long arg);
static int set_dma_mask(struct pci_dev * pdev);
static int init_pci_resource(bpOSD_COMP_EXTENSION pOSDExtension,
					struct pci_dev* pdev);
static void shasta_iounmap(bpOSD_COMP_EXTENSION p);
static int release_single_device(bpOSD_COMP_EXTENSION p);
static int st_init_common(
	bpOSD_COMP_EXTENSION *ppOSDExtension,
	struct Scsi_Host * shpnt,
	struct pci_dev *pdev,
	U32 re_size
	);

struct file_operations shasta_fops = {
	owner: THIS_MODULE,
	ioctl:		st_ioctl,
	fasync:		st_fasync,
	open:		st_open,
	release:		st_release
};

void Host_OSD_Delay(
	U32 DelayTime
	)
{
int i;
#define ST_UDELAY_MAX 1000
if(DelayTime <= ST_UDELAY_MAX)
	udelay(DelayTime);
else
	for(i = 0; i < DelayTime/ST_UDELAY_MAX; i++)
		udelay(ST_UDELAY_MAX);
}

U32 wrap_readl(void * mmiobase){
	return readl(mmiobase);
	}

U16 wrap_readw(void * mmiobase){
	return readw(mmiobase);
	}

U8 wrap_readb(void * mmiobase){
	return readb(mmiobase);
	}

void wrap_writel(void* mmiobase, U32 data){
	writel(data,mmiobase);
	}

void wrap_writew(void* mmiobase, U16 data){
	writew(data,mmiobase);
	}

void wrap_writeb(void* mmiobase, U8 data){
	writeb(data,mmiobase);
	}

unsigned char wrap_inb(int addr){
	return inb(addr);
	}

unsigned short wrap_inw(int addr){
	return inw(addr);
	}

unsigned int wrap_inl(int addr){
	return inl(addr);
	}

void wrap_outb(int addr,unsigned char b){
	outb(b,addr);
	}

void wrap_outw( int addr,unsigned short b){
	outw(b,addr);
	}

void wrap_outl( int addr,unsigned int b){
	outl(b,addr);
	}

void wrap_memcpyfromio(void* mmiobase, const void* buffer, unsigned long count){
memcpy_fromio(buffer, mmiobase, count);

}

void wrap_memcpytoio(void* mmiobase, const void* buffer, unsigned long count){
memcpy_toio(mmiobase, buffer, count);
}

int wrap_print(const char *fmt,...)
{	
#if VDEBUG
	static char buf[100];
	va_list args;
	int i;
	va_start(args,fmt);
	i=vsprintf(buf,fmt,args);
	i=printk(buf);

	va_end(args);
	return i;
#else
	return 0;
#endif
}

void 
init_data_pool(bpVOID OSDExtension,bpMEMORY_AVAILABLE pMemAvail)
{
	bpOSD_COMP_EXTENSION p = (bpOSD_COMP_EXTENSION)OSDExtension;
	//ULONG constantSize = FIELD_OFFSET(OSD_COMP_EXTENSION, TimerPoolBuffer);
	int i;

	for (i=0;i<MAX_DATA_BUFFER_COUNT;i++)
	{
		p->DataPool[i].VBuffer = (PU8)( pMemAvail->DMASafeAddress + i*( MAX_COPY_SIZE));
		p->DataPool[i].VDataBuffer = (PU8)( pMemAvail->DMASafeAddress 
			+ i*( MAX_COPY_SIZE) + sizeof(SRB_IO_CONTROL) + sizeof(MY_SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER));
		p->DataPool[i].PhDataBuffer = ( pMemAvail->DMASafePhysicalAddress.Value 
			+ i*( MAX_COPY_SIZE) + sizeof(SRB_IO_CONTROL) + sizeof(MY_SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER));
		//OSD_DPRINT( ("Buffer %d address is %p\n", i, p->DataPool[i].Buffer) );
		if(MAX_DATA_BUFFER_COUNT -1 == i)
			p->DataPool[i].next=NULL;
		else
			p->DataPool[i].next=&(p->DataPool[i+1]);
	}	
	p->pDataPoolNext = p->DataPool;
}

static PDATA_POOL 
alloc_data_pool(bpVOID OSDExtension)
{
	bpOSD_COMP_EXTENSION p = (bpOSD_COMP_EXTENSION)OSDExtension;
	unsigned long flags;
	PDATA_POOL pDataPool;
	spin_lock_irqsave(&p->datapool_lock, flags);
	pDataPool = p->pDataPoolNext;

	if(NULL == pDataPool){
		spin_unlock_irqrestore(&p->datapool_lock, flags);
		OSD_DPRINT( ("alloc_data_pool: data pool is empty\n") );
		return NULL;
	}	
	p->pDataPoolNext = pDataPool->next;
	pDataPool->next = NULL;
	spin_unlock_irqrestore(&p->datapool_lock, flags);
	return pDataPool;
}

static void 
release_data_pool(bpVOID OSDExtension, PDATA_POOL pDataPool)
{
	bpOSD_COMP_EXTENSION p = (bpOSD_COMP_EXTENSION)OSDExtension;
	unsigned long flags;

	spin_lock_irqsave(&p->datapool_lock, flags);
	if(NULL == pDataPool){
		spin_unlock_irqrestore(&p->datapool_lock, flags);
		OSD_DPRINT( ("release_data_pool: intend to put NULL data buffer to pool \n") );
		return;
	}	
	
		pDataPool->next = p->pDataPoolNext;
		p->pDataPoolNext = pDataPool;
		
	
	spin_unlock_irqrestore(&p->datapool_lock, flags);
}

void init_req_pool(bpOSD_COMP_EXTENSION p)
{
	int i;
	memset(p->ReqHeadPool, 0, ((sizeof(REQ_HEAD))*CAN_QUEUE));
	for(i=0; i<CAN_QUEUE; i++){
		if(CAN_QUEUE -1 == i){
			p->ReqHeadPool[i].next = NULL;
		}else{
			p->ReqHeadPool[i].next = &(p->ReqHeadPool[i+1]);
		}
		
	}

	p->pReqCurrent = p->ReqHeadPool;
}

static PREQ_HEAD get_req_from_pool(bpOSD_COMP_EXTENSION p)
{	
	PREQ_HEAD pReq;



	pReq = p->pReqCurrent;

	if(NULL == pReq){
		OSD_DPRINT( ("get_req_from_pool: pool is empty\n") );

		return NULL;
	}	

	p->pReqCurrent = pReq->next;
	
	pReq->next = (PREQ_HEAD)ST_REQ_IN_USE;
	pReq->pExt = (void *)p;
	init_timer(&pReq->req_timer);


	return pReq;
}

static void put_req_to_pool(bpOSD_COMP_EXTENSION p, PREQ_HEAD pReq)
{


	if(NULL == pReq || (pReq->next != (PREQ_HEAD)ST_REQ_IN_USE)){
		OSD_DPRINT( ("put_req_to_pool: invalid req\n") );

		return;
	}	
	memset(pReq, 0, sizeof(REQ_HEAD) );
	pReq->next = p->pReqCurrent;
	p->pReqCurrent = pReq;

}

static void st_add_req_timer(bpOSD_COMP_EXTENSION p,PREQ_HEAD pReq)
{

if(pReq->next == (PREQ_HEAD)ST_REQ_IN_USE &&
	!timer_pending(&pReq->req_timer)){
	pReq->retry_count ++;
	pReq->req_timer.expires=jiffies+1;
   	pReq->req_timer.data=(unsigned long)pReq;
   	pReq->req_timer.function=shasta_req_timer;
   	add_timer(&pReq->req_timer);
	}

}
/* function   :int st_open(struct inode *inodep, struct file *filep)
 * input        : Standard Parameters
 * output      : 0 if success, otherwise error reason
 * description  : open device
 */
static int st_open(struct inode *inodep, struct file *filep)
{

	unsigned int aptno = MINOR(inodep->i_rdev);
	bpOSD_COMP_EXTENSION p = CMPM_GetOSDExtensionFromDeviceExtension(gShasta_hostp[aptno]);	
		
	//OSD_DPRINT( ("try to open Adapter %d \n", aptno) );

	if(aptno >= MAX_ADAPTORS ||!gShasta_hostp[aptno] || gShastaMajor <= 0 || gShastaMajor != MAJOR(inodep->i_rdev) ){
	   	OSD_TRACE( ("open Adapter %d error\n", aptno) );
		return(-ENODEV);
	}
	filep->private_data = (void *)p;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_INC_USE_COUNT;
#endif
	return(0);
}

/* function   :int st_release(struct inode *inodep, struct file *filep)
 * input        : Standard Parameters
 * output      : 0 if success, otherwise error reason
 * description  : release device
 */
static int st_release(struct inode *inodep, struct file *filep)
{

	unsigned int aptno = MINOR(inodep->i_rdev);
	if(aptno >= MAX_ADAPTORS ||!gShasta_hostp[aptno] || gShastaMajor <= 0 || gShastaMajor != MAJOR(inodep->i_rdev))
		return(-ENODEV);


	st_fasync(-1, filep, 0);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	MOD_DEC_USE_COUNT;
#endif
	return(0);
}

/* function   : int st_ioctl(struct inode *inodep,
		struct file *filep,
		unsigned int req,
		unsigned long arg)
 * input        : Standard Parameters
 * output      : 0 if success, otherwise error reason
 * description  : device IOCTL entry
 */

static int st_ioctl(struct inode *inodep, struct file *filep, unsigned int req, unsigned long arg)
{
	int ret = 0;
	unsigned int aptno = MINOR(inodep->i_rdev);
	bpOSD_COMP_EXTENSION p = CMPM_GetOSDExtensionFromDeviceExtension(gShasta_hostp[aptno]);
	PSRB_IO_CONTROL pSrbCtl = (PSRB_IO_CONTROL)arg;	
	PIOCTL_DRIVERVERSION_FRAME pVersion;
	PREQ_HEAD pReq; 
	PSCSI_REQUEST_BLOCK Srb;
	PMY_SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER pSptd ;
	PVOID pMuExt;
	volatile int needsleep;
	PDATA_POOL pDataPool;
	void *usr_copy_start, *ker_copy_start;
	unsigned long copy_size;
	unsigned long flags;
	int error;
	/* Step1: basic parameter check */
	if(aptno >= MAX_ADAPTORS || !gShasta_hostp[aptno] || gShastaMajor<= 0 || gShastaMajor != MAJOR(inodep->i_rdev) )
		return(-ENODEV);
	
	if( !arg ) {
		return(0x10);
	}


	if( (pDataPool = alloc_data_pool(p) ) == NULL ){	
		return (-EIO);
		}	

	error = copy_from_user(pDataPool->VBuffer, (void *)arg, sizeof(SRB_IO_CONTROL));/*might sleep,must not lock*/

	if (error)
		goto err_out;

	pSrbCtl = (PSRB_IO_CONTROL)(pDataPool->VBuffer);	
	//OSD_TRACE( ("passthru ioctl number %x, req is %x \n",IOCTL_FASTTRAK_PASSTHRU, req) );
	/* Step2: Signature of the SRB_IO_CONTROL is wrong. Return immediately. */
	if(	(*((U32 *)&(pSrbCtl->Signature))!=FASTTRAK_IOCTL_SIGNATURE1) ||
		(*(((U32 *)&(pSrbCtl->Signature))+1)!=FASTTRAK_IOCTL_SIGNATURE2)
	  )
	 	goto err_out;
	
	usr_copy_start = (char *)arg + sizeof(SRB_IO_CONTROL);
	ker_copy_start = (void *)(pSrbCtl + 1);
	/*Step3: handle cmd */
	switch(req){
		case IOCTL_FASTTRAK_GETVERSION:
		
			pVersion = (PIOCTL_DRIVERVERSION_FRAME)ker_copy_start;
			pVersion->Major = VER_MAJOR;
			pVersion->Minor = VER_MINOR;
			pVersion->Oem = OEM;
			pVersion->BuildNo = BUILD_VER;
			pVersion->Signature[0] = 'NAPA';
			pVersion->ConsoleID = CONSOLEDEVICE_TARGETID;
			pVersion->HostNumber = p->host_no;
			copy_size = sizeof(IOCTL_DRIVERVERSION_FRAME);
	//		OSD_TRACE( ("ret = %x\n", ret) );
			break;
		
		case IOCTL_FASTTRAK_PASSTHRU:
		
			pSptd = (PMY_SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER)ker_copy_start;
			error = copy_from_user(ker_copy_start, usr_copy_start, sizeof(MY_SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER));
			if (error)
				goto err_out;
			/* Handle Console Request which related to the OSD. */ 
			if ( pSptd->TargetId==CONSOLEDEVICE_TARGETID && pSptd->Lun==0 ){
				ker_copy_start += sizeof(MY_SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER);
				usr_copy_start += sizeof(MY_SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER);
				if(pSptd->DataTransferLength > MAX_DATA_BUFFER_SIZE){
					printk("shasta: invalid ioctl buffer size\n");
					goto err_out;
				}
				if(pSptd->DataTransferLength){
				error = copy_from_user(ker_copy_start , usr_copy_start , 
					pSptd->DataTransferLength );

				if (error)
					goto err_out;
				}
				DRIVER_LOCK(p,flags);
				if( (pReq = get_req_from_pool(p) ) == NULL ){
					DRIVER_UNLOCK(p,flags);
					goto err_out;
				}							

				
				init_waitqueue_head(&(pReq->waitq) );	
				pReq->pSptd = (void *)pSptd;
				pReq->ReqType = PASSTHRU_REQ_HEAD_TYPE;
					
				Srb= &(pReq->Srb);
				
				Srb->TargetId = pSptd->TargetId;
				Srb->SenseInfoBufferLength = Min(pSptd->SenseInfoLength,SPT_SENSE_LENGTH);			
				Srb->Lun = pSptd->Lun;
				Srb->DataTransferLength = pSptd->DataTransferLength;
				
				copy_size = sizeof(MY_SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER) 
					+ pSptd->DataTransferLength;
				memcpy(Srb->Cdb, pSptd->Cdb, Min(16, pSptd->CdbLength) );
				//memcpy(pDataPool->VSenseBuffer, pSptd->SenseBuffer, Min(Srb->SenseInfoBufferLength,SPT_SENSE_LENGTH) );
						//OSD_DPRINT( ("Data Buffer Address is %p\n",pDataPool->Buffer) );
				pMuExt=CMPM_GetComponentExtension(p,COMP_TYPE_MU);	

				needsleep = 1; 	
				
				Srb->SenseInfoBuffer = ker_copy_start - Min(pSptd->SenseInfoLength,SPT_SENSE_LENGTH);
				Srb->DataBuffer = ker_copy_start;			
				if(0!=OSD_HandleSCSIRequest(p, Srb))
					needsleep = 0;

				if(needsleep){
					st_add_req_timer(p,pReq);
					}	
				DRIVER_UNLOCK(p,flags);

				if(needsleep){
					wait_event((pReq->waitq),!(pReq->ReqType) );
					}
				DRIVER_LOCK(p,flags);
				pSptd->SrbStatus = Srb->SrbStatus;
				pSptd->ScsiStatus = Srb->ScsiStatus;	
				put_req_to_pool(p, pReq);	
				DRIVER_UNLOCK(p,flags);
				ker_copy_start -= sizeof(MY_SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER);
				usr_copy_start -= sizeof(MY_SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER);

				if( (SRB_STATUS_SUCCESS == pSptd->SrbStatus) || (SRB_SEE_SENSE == pSptd->SrbStatus) )
					ret = 0;
				else
					ret = -EFAULT;

			} /* if console request. */
			else goto err_out;
			break;
		
		default:
			OSD_DPRINT( ("st_ioctl: unknown ioctl %x\n", req) );
			ret = -EINVAL;
			goto out;
	}	

	error = copy_to_user(usr_copy_start, ker_copy_start, copy_size);

	if (error )
		ret = -EFAULT;
out:
	release_data_pool(p, pDataPool);

	return ret;	
err_out:
	ret = -EFAULT;
	goto out;
}

static int st_fasync(int fd, struct file * filep, int mode)
{
	bpOSD_COMP_EXTENSION p = (bpOSD_COMP_EXTENSION)(filep->private_data);
	
	return fasync_helper(fd, filep, mode, &p->async_queue); 
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
typedef struct scsi_cmnd Scsi_Cmnd;
#endif
static char EndRequest(void *cmnd,void *fun,int result) 
{
	Scsi_Cmnd	     *cmd  = (Scsi_Cmnd *)cmnd ;
   	void	(*fn)(Scsi_Cmnd *) = (void *)fun ;
   	cmd->result = result;
   
   	fn(cmd);
   	return(0);
}

void
OSD_DumpSrb(
   IN PSCSI_REQUEST_BLOCK Srb
)
{
	OSD_TRACE(("Dump Srb...\n"));
	OSD_TRACE(("Lun=0x%x, TargetId=0x%x, Cdb[0]=0x%x.\n",
        Srb->Lun, Srb->TargetId, Srb->Cdb[0]));
}

void 
OSD_PrepareSG(
	IN bpOSD_COMP_EXTENSION pOSDExtension,
	IN PSCSI_REQUEST_BLOCK Srb,
	IN bpSGTABLE p_sgTable
	)
{
	PREQ_HEAD pReq = (PREQ_HEAD)Srb;
	Scsi_Cmnd* cmd = (Scsi_Cmnd *)(pReq->cmd);
	struct pci_dev *pDev = pOSDExtension->pdev;
	dma_addr_t dma_handle;
	struct scatterlist *sg; 
	int us;
SGTableStandardInitialize(p_sgTable);
	if(pReq->ReqType & PASSTHRU_REQ_HEAD_TYPE){
		PDATA_POOL pd;
		unsigned long flags;
		spin_lock_irqsave(&pOSDExtension->datapool_lock, flags);
		pd = pOSDExtension->DataPool;
		for(us=0;us<MAX_DATA_BUFFER_COUNT;us++,pd++){
			if(pd->VDataBuffer == Srb->DataBuffer){
				p_sgTable->table[0].StartAddress.Value = pd->PhDataBuffer;
				break;
				}
			}
		spin_unlock_irqrestore(&pOSDExtension->datapool_lock, flags);
		p_sgTable->SGCount = cpu_to_le16(1);
		p_sgTable->table[0].Count   = cpu_to_le32(Srb->DataTransferLength);
		p_sgTable->table[0].CtrlFlag = SG_CF_EOT|SG_CF_64B | SG_CF_HOST;
		p_sgTable->SizeInByte = cpu_to_le32(Srb->DataTransferLength);
		return;
		}
	else if (cmd->use_sg){
		sg = (struct scatterlist *) cmd->request_buffer;
		p_sgTable->SGCount = cpu_to_le16((U16)pci_map_sg(pDev,sg,cmd->use_sg,scsi_to_pci_dma_dir(cmd->sc_data_direction)));

		for (us = 0; us < p_sgTable->SGCount ; us++,sg++){
      			p_sgTable->table[us].Count  = cpu_to_le32((U32)sg_dma_len(sg));
      				

			p_sgTable->table[us].StartAddress.Value = cpu_to_le64(sg_dma_address(sg));


			p_sgTable->table[us].CtrlFlag = SG_CF_64B | SG_CF_HOST;
    		}
			p_sgTable->table[--us].CtrlFlag |= SG_CF_EOT;
			
	}else{
#if defined(CONFIG_X86_64)
#if defined(CONFIG_DISCONTIGMEM) && (LINUX_VERSION_CODE>=KERNEL_VERSION(2,4,18) )
		dma_handle = pci_map_page(pDev,virt_to_page(cmd->request_buffer),((unsigned long)cmd->request_buffer & ~PAGE_MASK),cmd->request_bufflen,scsi_to_pci_dma_dir(cmd->sc_data_direction));
#else
		dma_handle = pci_map_single(pDev,
						(cmd->request_buffer),
						cmd->request_bufflen,
						scsi_to_pci_dma_dir(cmd->sc_data_direction) );
#endif
#else
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,4,18) 
	  	dma_handle = pci_map_page(pDev,virt_to_page(cmd->request_buffer),((unsigned long)cmd->request_buffer & ~PAGE_MASK),cmd->request_bufflen,scsi_to_pci_dma_dir(cmd->sc_data_direction));
 #else
 		dma_handle = pci_map_single(pDev,
						(cmd->request_buffer),
						cmd->request_bufflen,
						scsi_to_pci_dma_dir(cmd->sc_data_direction) );
#endif
#endif
 		 p_sgTable->SGCount = cpu_to_le16(1);

		p_sgTable->table[0].StartAddress.Value = cpu_to_le64(dma_handle);

		p_sgTable->table[0].Count   = cpu_to_le32((U32) cmd->request_bufflen);
		p_sgTable->table[0].CtrlFlag = SG_CF_EOT|SG_CF_64B | SG_CF_HOST;
		
		memcpy(&pReq->sg, p_sgTable->table, sizeof(SG));/*copy one sg for later unmapping*/
  	}
p_sgTable->SizeInByte = cpu_to_le32(cmd->request_bufflen);
}

void OSD_NotifyBusChange(
	IN PVOID ComponentExtension
)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
	/*bpOSD_COMP_EXTENSION p = (bpOSD_COMP_EXTENSION)
			CMPM_GetComponentExtension(ComponentExtension, COMP_TYPE_OSD);
	
	if(p&&p->State != SYS_STATE_FULLY_STARTED)
		return;
	scsi_scan_host(pOSDExtension->host);*/
#endif
	
}

void OSD_NotifyEvent(
	IN PVOID ComponentExtension
)
{
	bpOSD_COMP_EXTENSION pOSDExtension = (bpOSD_COMP_EXTENSION)
			CMPM_GetComponentExtension(ComponentExtension, COMP_TYPE_OSD);

	OSD_TRACE(("There is an event. \n"));
	if(pOSDExtension->async_queue)
		kill_fasync(&pOSDExtension->async_queue, SIGIO, POLL_OUT);
	
}

void OSD_GetCurrentTime(PU64 time)
{
struct timeval	tv;
do_gettimeofday(&tv);

time->Value = cpu_to_le64(tv.tv_sec); 
}

static void st_set_cmdresult(bpOSD_COMP_EXTENSION p,PSCSI_REQUEST_BLOCK Srb)
{
	REQ_HEAD *pReq = (PREQ_HEAD)Srb;
   	Scsi_Cmnd 		*cmd; 
   	void		(*fn)(void *);
    	dma_addr_t dma_handle;	
	int	result=0; 
	cmd = (Scsi_Cmnd *)(pReq->cmd);
	fn = pReq->fn;

	if(scsi_to_pci_dma_dir(cmd->sc_data_direction)!=PCI_DMA_NONE){
 		if (cmd->use_sg){
			pci_unmap_sg(p->pdev,cmd->request_buffer,cmd->use_sg,scsi_to_pci_dma_dir(cmd->sc_data_direction));			
		}
 		else{
 			dma_handle = (dma_addr_t)le64_to_cpu(pReq->sg.StartAddress.Value); 
 #if defined(CONFIG_X86_64)
#if defined(CONFIG_DISCONTIGMEM) && (LINUX_VERSION_CODE>=KERNEL_VERSION(2,4,18) )
			pci_unmap_page(p->pdev,dma_handle,cmd->request_bufflen,scsi_to_pci_dma_dir(cmd->sc_data_direction));
#else
			pci_unmap_single(p->pdev,dma_handle,cmd->request_bufflen,scsi_to_pci_dma_dir(cmd->sc_data_direction));
#endif
#else
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,4,18) 
	  		pci_unmap_page(p->pdev,dma_handle,cmd->request_bufflen,scsi_to_pci_dma_dir(cmd->sc_data_direction));
#else
 			pci_unmap_single(p->pdev,dma_handle,cmd->request_bufflen,scsi_to_pci_dma_dir(cmd->sc_data_direction));

#endif
#endif
 			}
  	}
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	if(cmd->cmnd[0] == READ_CAPACITY)
	{
		if(*(U32 *)cmd->request_buffer==0xffffffff)
			((U8 *)cmd->request_buffer)[3] = 0xfe;
	}
#endif
	switch(Srb->SrbStatus)
	{
		case SRB_STATUS_SUCCESS:
			result= DID_OK << 16 | COMMAND_COMPLETE << 8;
			break;
		case SRB_STATUS_SELECTION_TIMEOUT:
			result= DID_BAD_TARGET << 16 | COMMAND_COMPLETE << 8;
			break;
		case SRB_SEE_SENSE:
			result= (DRIVER_SENSE << 24) | (DID_OK << 16) | (CHECK_CONDITION << 1) ;
			break;
		case SRB_STATUS_BUSY:
			result = DID_NO_CONNECT << 16 | COMMAND_COMPLETE << 8;
			break;

		default:
			result = DID_ERROR << 16 | COMMAND_COMPLETE << 8; 
			break;
  	}

	put_req_to_pool(p, pReq);
   	
	EndRequest(cmd,fn,result);
}

void OSD_TrueRAIDReturn(
	IN PVOID OSDExt,
	IN PSCSI_REQUEST_BLOCK Srb
	)
{
	bpOSD_COMP_EXTENSION p = (bpOSD_COMP_EXTENSION)OSDExt;
	REQ_HEAD *pReq = (PREQ_HEAD)Srb;
 
	OSD_TRACE( ("RAID_Return: Req = 0x%p, SrbStatus =%x, ScsiStatus = %x, ReqType=%x\n", p, Srb->SrbStatus, Srb->ScsiStatus, pReq->ReqType) );
	
	if(timer_pending(&pReq->req_timer))
		del_timer(&pReq->req_timer);
	
	if(p->pReqWaiting==pReq)
		p->pReqWaiting =  NULL;
	
	if(pReq->ReqType & PASSTHRU_REQ_HEAD_TYPE){
	
		pReq->ReqType = 0;
		if(waitqueue_active(&pReq->waitq)){	
			wake_up(&(pReq->waitq) );
			}
		return;
	}
  	//need improved
  	st_set_cmdresult(p, Srb);

}
#define SH_POLL_TIMES 100000
#define SH_POLL_USEC 10
/* function   : static int shasta_abort(Scsi_Cmnd *cmd);
 * input        : cmd = SCSI command to be aborted
 * output      : SUCCESS
 * description  : SCSI command abort function
 */
static int shasta_abort(Scsi_Cmnd *cmd)
{
	bpOSD_COMP_EXTENSION p;	
	struct Scsi_Host *phost;
	int i;
	unsigned long smpflag;
	bpMU_COMP_EXTENSION pMUExtension;
	PTR_SIZE BaseAddress;
	U32 Data;
	OSD_DPRINT( ("shasta_abort: returning SUCCESS\n") );
	if (!cmd) {
		printk(KERN_WARNING "shasta: shasta_abort(): Invalid Scsi_Cmnd\n");
		return (FAILED);
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
 	phost = cmd->host;
#else
	phost = cmd->device->host;
#endif
	p = (bpOSD_COMP_EXTENSION)phost->hostdata[0];
	if (p == NULL) {
		printk(KERN_WARNING "shasta: shasta_abort(): Invalid OSD extension\n");
		return (FAILED);
	}
	
	pMUExtension = (bpMU_COMP_EXTENSION)
					CMPM_GetComponentExtension(p, COMP_TYPE_MU);
	BaseAddress = (PTR_SIZE)pMUExtension->BaseAddress;	

	DRIVER_LOCK(p,smpflag);
	for(i=0;i<CAN_QUEUE;i++)
		if(p->ReqHeadPool[i].cmd == cmd){
			
			p->pReqWaiting = &(p->ReqHeadPool[i]);
			if(timer_pending(&(p->ReqHeadPool[i].req_timer)))
				del_timer(&(p->ReqHeadPool[i].req_timer));
			put_req_to_pool(p, &(p->ReqHeadPool[i]));
			break;
			
			}
	if(i >= CAN_QUEUE){ 
		DRIVER_UNLOCK(p,smpflag);
		return FAILED;
		}
	OSDI_SetRestartFlag(p);

	for(i=0;i<SH_POLL_TIMES;i++){
	/* Let me check the Outbound Doorbell Register */
	Data = OSD_HwReadU32( BaseAddress+MU_OUTBOUND_DOORBELL );
	if(Data == 0){
		p->pReqWaiting = NULL;
		DRIVER_UNLOCK(p,smpflag);	
		return FAILED;
		}
	/* Let me clear the interrupt. */
	OSD_HwWriteU32( BaseAddress+MU_OUTBOUND_DOORBELL, Data );
	if ( Data&MU_OUTBOUND_DOORBELL_HASEVENT )
		{
		Data &=~MU_OUTBOUND_DOORBELL_HASEVENT;
		OSD_NotifyEvent(p);
		}
	/* Bus Change Notification from Firmware. */
	if ( Data&MU_OUTBOUND_DOORBELL_BUSCHANGE )
		{
		Data &= ~MU_OUTBOUND_DOORBELL_BUSCHANGE;
		schedule_work(&p->work_q);
		}
	/* It's MU Interrupt. */
	if ( Data!=0 )
		MU_HandleInterrupt( pMUExtension, Data );

	
	if(p->pReqWaiting == NULL)
		break;
	udelay(SH_POLL_USEC);
	}
	p->pReqWaiting = NULL;
	DRIVER_UNLOCK(p,smpflag);	
		
	return i<SH_POLL_TIMES? SUCCESS : FAILED;
}

/* function   : static int shasta_reset(Scsi_Cmnd *cmd)
 * input        : Standard Parameters
 * output      : SUCCESS
 * description  : SCSI reset function
 */
static int shasta_reset(Scsi_Cmnd *cmd)
{
	OSD_DPRINT( ("shasta_reset: returning SCSI_RESET_SUCCESS\n") );
	if (!cmd) {
		printk(KERN_WARNING "shasta: shasta_reset(): Invalid Scsi_Cmnd\n");
		return (FAILED);
	}

   	return SUCCESS;
}

/* function   : void shasta_intr(int irq, void *dev_id, struct pt_regs *regs);
 * input        : Standard Parameters
 * output      : NONE
 * description  : interrupt handler
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static void shasta_intr(int irq, void *dev_id, void *regs)
#else
static irqreturn_t shasta_intr(int irq, void *dev_id, void *regs)
#endif
{	
	bpOSD_COMP_EXTENSION p = (bpOSD_COMP_EXTENSION)dev_id; 
	bpMU_COMP_EXTENSION pMUExtension;	
	PTR_SIZE BaseAddress;
	U32 Data;
	
	unsigned long smpflag;
	
	if(!dev_id)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)		
		return;
#else
		return (IRQ_NONE);
#endif
	
	if(p->irq != irq)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)		
		return;
#else
		return (IRQ_NONE);
#endif
	
	pMUExtension = (bpMU_COMP_EXTENSION)
					CMPM_GetComponentExtension(p, COMP_TYPE_MU);
	BaseAddress = (PTR_SIZE)pMUExtension->BaseAddress;	
	OSD_DPRINT( ("shasta_intr: \n") );
	DRIVER_LOCK(p,smpflag);
	/* Let me check the Outbound Doorbell Register */
	Data = OSD_HwReadU32( BaseAddress+MU_OUTBOUND_DOORBELL );
	if(Data == 0){
		DRIVER_UNLOCK(p,smpflag);	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)		
		return;
#else
		return (IRQ_NONE);
#endif
		}
	/* Let me clear the interrupt. */
	OSD_HwWriteU32( BaseAddress+MU_OUTBOUND_DOORBELL, Data );
	//OSD_ASSERT( OSD_HwReadU32( BaseAddress+MU_OUTBOUND_DOORBELL )==0 );
	//OSD_TRACE(("MU_OUTBOUND_DOORBELL=0x%x(0x4=StatusHead, 0x2=RequestTail).\n", Data));
	if ( Data&MU_OUTBOUND_DOORBELL_HASEVENT )
	{
		Data &=~MU_OUTBOUND_DOORBELL_HASEVENT;
		OSD_NotifyEvent(p);
	}	
	/* Bus Change Notification from Firmware. */
	if ( Data&MU_OUTBOUND_DOORBELL_BUSCHANGE )
	{
		Data &= ~MU_OUTBOUND_DOORBELL_BUSCHANGE;
		schedule_work(&p->work_q);
	}

	/* It's MU Interrupt. */
	if ( Data!=0 )
		MU_HandleInterrupt( pMUExtension, Data );


	DRIVER_UNLOCK(p,smpflag);	
	
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)		
	return (IRQ_HANDLED);
#endif
}

static void shasta_req_timer(unsigned long arg)
{
	PREQ_HEAD pReq = (PREQ_HEAD)arg;
	bpOSD_COMP_EXTENSION p;
	bpMU_COMP_EXTENSION pMUExtension;	
	PTR_SIZE BaseAddress;
	U32 Data;
	unsigned long smpflag;
	if(!pReq) return;
	p = (bpOSD_COMP_EXTENSION)pReq->pExt;
	if(!p) return;
	
	DRIVER_LOCK(p,smpflag);
	if(!OSDI_IsRestart(p)&&!(pReq->ReqType & PASSTHRU_REQ_HEAD_TYPE)){
		DRIVER_UNLOCK(p,smpflag);		
		return;
		}
	if((pReq->next != (PREQ_HEAD)ST_REQ_IN_USE)||timer_pending(&pReq->req_timer)){
		DRIVER_UNLOCK(p,smpflag);		
		return;
		}
	pMUExtension = (bpMU_COMP_EXTENSION)
					CMPM_GetComponentExtension(p, COMP_TYPE_MU);
	BaseAddress = (PTR_SIZE)pMUExtension->BaseAddress;	

	/* Let me check the Outbound Doorbell Register */
	Data = OSD_HwReadU32( BaseAddress+MU_OUTBOUND_DOORBELL );
	if(Data == 0){
		if((pReq->ReqType & PASSTHRU_REQ_HEAD_TYPE)&&
			pReq->retry_count >= ST_MAX_RETRY_COUNT){
			pReq->ReqType = 0;
			if(waitqueue_active(&pReq->waitq)){	
				wake_up(&(pReq->waitq) );
				}
			}
		else
			st_add_req_timer(p,pReq);
		DRIVER_UNLOCK(p,smpflag);		
		return;
		}
	p->pReqWaiting = pReq;
	/* Let me clear the interrupt. */
	OSD_HwWriteU32( BaseAddress+MU_OUTBOUND_DOORBELL, Data );
	if ( Data&MU_OUTBOUND_DOORBELL_HASEVENT )
	{
		Data &=~MU_OUTBOUND_DOORBELL_HASEVENT;
		OSD_NotifyEvent(p);
	}
	/* Bus Change Notification from Firmware. */
	if ( Data&MU_OUTBOUND_DOORBELL_BUSCHANGE )
	{
		Data &= ~MU_OUTBOUND_DOORBELL_BUSCHANGE;
		schedule_work(&p->work_q);
	}
	/* It's MU Interrupt. */
	if (Data!=0 )
		MU_HandleInterrupt( pMUExtension,Data);

	if(p->pReqWaiting != NULL){
		if((pReq->ReqType & PASSTHRU_REQ_HEAD_TYPE)&&
			pReq->retry_count >= ST_MAX_RETRY_COUNT){
			pReq->ReqType = 0;
			if(waitqueue_active(&pReq->waitq)){	
				wake_up(&(pReq->waitq) );
				}
			}
		else
			st_add_req_timer(p,pReq);
		}
	p->pReqWaiting = NULL;	
	DRIVER_UNLOCK(p,smpflag);	
}	

/* function   : static int shasta_queue(Scsi_Cmnd *cmd, void (*fn)(Scsi_Cmnd *))
 * input        : cmd = SCSI command
 *		     : fn = command completion function
 * output      : 0
 * description  : SCSI queuecommand function
 */

static int
shasta_queue(Scsi_Cmnd *cmd, void (*fn)(Scsi_Cmnd *))
{
	bpOSD_COMP_EXTENSION p=NULL;
	REQ_HEAD		*pReq;
	int us;
 	struct Scsi_Host *phost;
	unsigned int channel,id,lun;
	unsigned long flags;
	
	OSD_TRACE( ("Enter Shasta_queue\n") );	
	
	/* Step1: get extension pointer and do some security check */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
 	phost = cmd->host;
	channel = cmd->channel;
	id = cmd->target;
	lun = cmd->lun;
#else
	phost = cmd->device->host;
	channel = cmd->device->channel;
	id = cmd->device->id;
	lun = cmd->device->lun;
#endif

	if(cmd->cmnd[0] == INQUIRY){/* work around for firmware lun issue--edl*/
		if(lun > gSt_max_lun)
			gSt_max_lun = lun;		
		}

	if(gSt_max_lun == 0){
		if(channel > 0){
			lun = channel;
			channel = 0;
			}				
		}
	
	if(!phost || channel != 0 || id >= (OSDM_MAX_TARGET_NUM) || lun >= OSDM_MAX_LUN_PER_TARGET) {
    		cmd->result = (DID_BAD_TARGET << 16);
    		fn(cmd);
    		return 0;
 	 }
	
	p = (bpOSD_COMP_EXTENSION)phost->hostdata[0];	
	/*Step2: get req and initialize req */
	
	spin_lock(&p->timer_lock);
	if( (pReq = get_req_from_pool(p) )== NULL){
		
		spin_unlock(&p->timer_lock);
		cmd->result = (DID_BUS_BUSY << 16);
		fn(cmd);
		return(0);
	}	
	pReq->cmd                 = (void *)cmd ;
   	pReq->fn                  = (void *)fn  ;
	pReq->Srb.SrbStatus 	= 1 ;
   	pReq->Srb.ScsiStatus  	= 0 ;
   	pReq->Srb.TargetId	= id ;
   	pReq->Srb.Lun    	= lun ;
   	pReq->Srb.SenseInfoBufferLength = SCSI_SENSE_BUFFERSIZE;
   	pReq->Srb.DataTransferLength = cmd->request_bufflen;
 	pReq->Srb.DataBuffer = (void *)cmd->request_buffer ;
       pReq->Srb.SenseInfoBuffer= (void *)cmd->sense_buffer ;

        OSD_TRACE( ("shasta_queue: req = 0x%p, ReqType = 0x%x, target=0x%x, cdb[0]=0x%x\n", pReq, pReq->ReqType, id, cmd->cmnd[0]) );

   	for( us=0 ; us < MAX_COMMAND_SIZE ; us++ ) {
	   	pReq->Srb.Cdb[us] = cmd->cmnd[us] ;
   	}
	/*Step3: if None R/W cmd, set DataBuffer and DataTransferLengh 0 */
 	if((cmd->sc_data_direction) == PCI_DMA_NONE){
 		pReq->Srb.DataTransferLength = 0;
 		pReq->Srb.DataBuffer = NULL ;
  	}

	/*Step4: send cmd to MU */
	if(0!=OSD_HandleSCSIRequest(p, (PSCSI_REQUEST_BLOCK)pReq)){
		st_set_cmdresult(p,&pReq->Srb);
		spin_unlock(&p->timer_lock);
		return 0;
		}

	if(OSDI_IsRestart(p)){
		st_add_req_timer(p,pReq);
		}
	spin_unlock(&p->timer_lock);
	OSD_TRACE( ("Leave shasta_queue\n") );
	
    	return 0;
}

static int set_dma_mask(struct pci_dev * pdev)
{
gHighmem_io = 0;
#if defined(_64BPLATFORM) || defined(CONFIG_HIGHMEM64G)
		/* Update our PCI device dma_mask for full 64 bits */
		if (pci_set_dma_mask(pdev, 0xffffffffffffffffULL) == 0) {
			
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
			if (pci_set_consistent_dma_mask(pdev,
			    0xffffffffffffffffULL)) {
				OSD_DPRINT( ("Shasta: Failed to set 64 bit PCI consistent mask, using 32 bit.\n") );
				return(pci_set_consistent_dma_mask(pdev,
				    0xffffffffULL) );
			}
#endif
			gHighmem_io = 1;

			return 0;
		} else {
			OSD_DPRINT( ("Shasta: Failed to set 64 bit PCI DMA mask, falling back to 32 bit MASK.\n") );
			return(pci_set_dma_mask(pdev, 0xffffffffULL) );
		}
#else		
	return(pci_set_dma_mask(pdev, 0xffffffffULL) );
#endif
}


static int init_pci_resource(bpOSD_COMP_EXTENSION p,
					struct pci_dev* pdev)
{
	int i;
	p->bitmap_ioremap = 0;
	for(i=0; i<BASE_ADDRESS_MAX_NUM; i++){
		if(pci_resource_flags(pdev, i) & IORESOURCE_MEM) {
			if(!(p->BaseAddress[i] = (LPVOID)ioremap(pci_resource_start(pdev, i),
		            pci_resource_len(pdev, i) ))){
		            
		            return -ENOMEM;
				}
			p->bitmap_ioremap |= (1<<i);
			OSD_TRACE( ("BA[%d]=%p\n memmap",i,p->BaseAddress[i] ) );
			}
		else{
			p->BaseAddress[i] = (LPVOID) pci_resource_start(pdev, i);
			OSD_TRACE( ("BA[%d]=%p\n iomap",i,p->BaseAddress[i] ) );
			}
	}
	return 0;

}

static void shasta_iounmap(bpOSD_COMP_EXTENSION p)
{
int i;
for(i = 0; i < BASE_ADDRESS_MAX_NUM; i++)
	if(p->bitmap_ioremap & (1<<i))
		iounmap(p->BaseAddress[i]);
p->bitmap_ioremap = 0;
}

static int release_single_device(bpOSD_COMP_EXTENSION p)
{
	int i;
	unsigned long flags;
	PREQ_HEAD pReq;
	PREQ_HEAD pReqKM = NULL;
/* Step1: send req to stop bga */
	for(i=0; i<CONSOLEDEVICE_TARGETID*2; i++){
		DRIVER_LOCK(p,flags);
		if( (pReq = get_req_from_pool(p) )== NULL){
			DRIVER_UNLOCK(p,flags);
			if( (pReq = pReqKM = kmalloc(sizeof(REQ_HEAD),GFP_KERNEL) )== NULL){
				break;
			}
			else pReq->next = (PREQ_HEAD)ST_REQ_IN_USE;
		}		
		pReq->Srb.SrbStatus 	= 1 ;
   		pReq->Srb.ScsiStatus  	= 0 ;
		
		memset(pReq->Srb.Cdb,0,MAX_COMMAND_SIZE);
		if(i<CONSOLEDEVICE_TARGETID){
			pReq->Srb.TargetId     	= i;
   			pReq->Srb.Lun     = 0 ;		
   			pReq->Srb.Cdb[0] = scsiSYNCHRONIZE_CACHE;
			}
		else{
   			pReq->Srb.TargetId    	= i-CONSOLEDEVICE_TARGETID;
   			pReq->Srb.Lun     = 0 ;
			pReq->Srb.Cdb[0] = 0xe1; /*svucCONTROLLER_COMMAND;*/
			pReq->Srb.Cdb[1] = 0x0e; /*ccmdPOWER_STATE_CHANGING;*/
			pReq->Srb.Cdb[2] = 0x01; /*CCMD_POWER_SAVING;*/
			}
		pReq->Srb.DataTransferLength = 0;
 		pReq->Srb.DataBuffer = NULL;
 
	 	pReq->Srb.SenseInfoBufferLength = SCSI_SENSE_BUFFERSIZE ;
       	pReq->Srb.SenseInfoBuffer = p->SenseBuffer;
		
		pReq->ReqType = PASSTHRU_REQ_HEAD_TYPE;
   		
   		init_waitqueue_head(&pReq->waitq);
		OSD_HandleSCSIRequest(p, (PSCSI_REQUEST_BLOCK)pReq);
		st_add_req_timer(p,pReq);
   		DRIVER_UNLOCK(p,flags);
		wait_event(pReq->waitq,!pReq->ReqType);

		if(pReqKM){
			kfree(pReqKM);
			pReqKM = NULL;
			}
		else	{	
			DRIVER_LOCK(p,flags);
			put_req_to_pool(p, pReq);
			DRIVER_UNLOCK(p,flags);
			}
		
	}
	
	CMPM_StopComponents(p);

	free_irq(p->irq, (void *)p);
	
	for(i=0;i<CAN_QUEUE;i++){
		if(timer_pending(&p->ReqHeadPool[i].req_timer))
			del_timer(&p->ReqHeadPool[i].req_timer);
	}
	
	shasta_iounmap(p);
	pci_release_regions(p->pdev);
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)	
	scsi_host_put(p->host);
#endif

	pci_disable_device(p->pdev);	
	pci_set_drvdata(p->pdev, NULL);

	if(p->DMASafeAddress) 
		pci_free_consistent(p->pdev,p->DMASafeSize,(void *)p->DMASafeAddress,p->dma_handle);
	
	for(i=0; i < gNumAdapter; i++){
		
		if(gShasta_hostp[i] && (gShasta_hostp[i] == p->DeviceExtension) ){	
			vfree(p->DeviceExtension);	
			OSD_TRACE( ("Clear hostp in Adapter id %d hostp 0x%p, extension 0x%p\n", i, gShasta_hostp[i], p) );
			gShasta_hostp[i] = NULL;
			break;
		}
	}	

	return 1;
	
}

static int st_init_common(
	bpOSD_COMP_EXTENSION *ppOSDExtension,
	struct Scsi_Host * shpnt,
	struct pci_dev *pdev,
	U32 re_size
	)
{
	bpOSD_COMP_EXTENSION pOSDExtension;
	shpnt->max_channel = OSDM_MAX_LUN_PER_TARGET;/*from 0 to MAX_LUN-1,for lun issue in firmware--edl*/
    	shpnt->max_id = OSDM_MAX_TARGET_NUM;
    	shpnt->max_lun = OSDM_MAX_LUN_PER_TARGET;/*from 0 to MAX_LUN-1--edl*/
	shpnt->unique_id = pdev->irq;
	shpnt->max_cmd_len = MAX_COMMAND_SIZE;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,4) && \
	LINUX_VERSION_CODE  < KERNEL_VERSION(2,5,0)
	scsi_set_pci_device(shpnt, pdev);
#endif
	/*Step 3: get cached memory resource for extension*/
    	gShasta_hostp[gNumAdapter] = (unsigned char *) vmalloc(re_size);
	if(NULL == gShasta_hostp[gNumAdapter]){
		OSD_DPRINT( ("Shasta: unable to allocate cached memory 0x%x for extension. \n", re_size) );
		return -ENOMEM;
	}	
    	memset(gShasta_hostp[gNumAdapter], 0, re_size);    	
	pOSDExtension=*ppOSDExtension=CMPM_GetOSDExtensionFromDeviceExtension(gShasta_hostp[gNumAdapter]);

	pOSDExtension->host = shpnt;
	pOSDExtension->host_no = shpnt->host_no;
	pOSDExtension->AdapterIndex = gNumAdapter;
	pOSDExtension->DeviceExtension = (bpVOID)(gShasta_hostp[gNumAdapter]);	
	pOSDExtension->bus_num=pdev->bus->number;
	pOSDExtension->devfn=pdev->devfn;
	pOSDExtension->pdev = pdev;
	pOSDExtension->present = 1;
	pOSDExtension->irq = pdev->irq;
	shpnt->hostdata[0] = (unsigned long)pOSDExtension;
  	//pOSDExtension->major = gShastaMajor;

	pci_set_drvdata(pdev, pOSDExtension);
	INIT_WORK(&pOSDExtension->work_q, OSD_NotifyBusChange, (void *)pOSDExtension);
  	/*Step 4 get uncached memory */
	pOSDExtension->ControllerInfo.DeviceID = (pdev->vendor << 16) | pdev->device;
	pOSDExtension->ControllerInfo.SubDeviceID =(pdev->subsystem_vendor << 16) | pdev->subsystem_device;
	pOSDExtension->ControllerInfo.SystemIoBusNumber = pdev->bus->number;
	pOSDExtension->ControllerInfo.SlotNumber = pdev->devfn;
	pOSDExtension->ControllerInfo.InterruptVector = pdev->irq;
	return 0;

}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)

/* function   : int shasta_proc_info(char *buffer, char **start, off_t offset,
			int length, int hostno, int func);
 * input        : Standard Parameters
 * output      : byte count printed
 * description  : PROC filesystem entry
 */
static int shasta_proc_info(char *buffer, char **start, off_t offset,
			int length, int hostno, int func)
{
	int len;

	*start = buffer;
	if (func == TRUE) {
		/*
		 * The user has done a write and wants us to take the
		 * data in the buffer and do something with it.
		 * proc_scsiwrite calls us with inout = 1
		 *
		 * Read data from buffer (writing to us) - NOT SUPPORTED
		 */
		return -EINVAL;
	}

	len = sprintf(buffer,"***** SuperTrak EX8350/8300 Driver %s %s*****\n", SHASTA_DRIVER_VERSION, SHASTA_DRIVER_DATE);

	len += sprintf(buffer + len,"***** Copyright 2005 by Promise Technology, Inc.   *****\n\n");

	*(buffer + len) = '\0';
 
	*start = buffer + (offset );	/* Start of wanted data */
	len -= (offset );
	if(len > length) {
		len = length;
	} else if(len < 0){
		len = 0;
		**start = '\0';
	}
	return len;
}	

/* function   : static int shasta_biosparam
 * input        : Standard Parameters
 * output      : 0
 * description  : SCSI biosparam function
 */
static int shasta_biosparam(Disk *disk, kdev_t dev, int Geometry[])
{
 if (disk->capacity > 0x200000) {	
  	Geometry[0]=255;	
  	Geometry[1]=63;  	
  	Geometry[2]=disk->capacity/(Geometry[0]*Geometry[1]);
 
	// capacity larger than 526GB(C/H/S = 65535/255/63)
  	/*if (Geometry[2] > 65535) {
   	   Geometry[2] = 65535;
   	   Geometry[1] = disk->capacity / (Geometry[2]*Geometry[0]);
  	} */ 
   }
   else{
   	Geometry[0]=64;  		
   	Geometry[1]=32;  		
   	Geometry[2]=disk->capacity / (Geometry[0]*Geometry[1]);  
   	}
 
   return 0;
}
	
/* function   : static int shasta_notifier_func(struct notifier_block *pnb,unsigned long code, void *unused)
 * input        : Standard Parameters
 * output      : NOTIFY_DONE
 * description  : stop bga and enclosure, synchronize cache
 */
static int shasta_notifier_func(struct notifier_block *pnb,unsigned long code, void *unused)
{
	bpOSD_COMP_EXTENSION p;
	int i;

	for(i=0; i < gNumAdapter; i++){
		
		if(gShasta_hostp[i]){	
			p = CMPM_GetOSDExtensionFromDeviceExtension(gShasta_hostp[i]);
			release_single_device(p);
		}
	}		
	
	return(NOTIFY_DONE);
}

static struct notifier_block shasta_notifier={
	shasta_notifier_func, NULL, 0
};

static void shasta_ha_init(struct pci_dev *pdev, Scsi_Host_Template *template)
{
	bpOSD_COMP_EXTENSION pOSDExtension;
	u32 rc = PCIBIOS_SUCCESSFUL;
	struct Scsi_Host * shpnt;
	ULONG DMAMemoryRequirement = 0;
	U32 re_size;	
	re_size = CMPM_GetComponentExtensionRequirement(gIsHibernation);

	if(gNumAdapter >= MAX_ADAPTORS){
		printk(KERN_WARNING
			"Too many cards present\n");
			return;
		}	
	/*Step1: enable and initialize pci device*/
	if ((rc = pci_enable_device(pdev))) {
		printk(KERN_WARNING
			"Cannot enable adapter\n");
		return;
	}

	rc = pci_request_regions(pdev, SHASTA_DEV_NAME);
	if (rc)
		goto err_out;

	rc = set_dma_mask(pdev);
	if (rc)
		goto err_out_regions;
	template->highmem_io &= gHighmem_io;

	pci_set_master(pdev);
		
	/*Step2: alloc and initialize scsi host */
	shpnt = scsi_register(template, sizeof(void *));

	if (!shpnt) {
		printk(KERN_WARNING
			"call to scsi_register failed!\n");
		rc = -ENOMEM;
		goto err_out_regions;
	}
	if(st_init_common(&pOSDExtension, shpnt, pdev, re_size)){
		rc = -ENOMEM;
		goto err_out_host;
		}

	DMAMemoryRequirement = 
		CMPM_GetUncachedMemoryRequirement(pOSDExtension, gIsHibernation);
	
	pOSDExtension->DMASafeAddress = pci_alloc_consistent(pdev,DMAMemoryRequirement,&pOSDExtension->dma_handle);
	if(NULL == pOSDExtension->DMASafeAddress){
		OSD_DPRINT( ("Shasta: unable to allocate uncached memory 0x%x\n", DMAMemoryRequirement) );
		rc = -ENOMEM;
		goto err_out_vmalloc;
	}
	pOSDExtension->DMASafeSize = DMAMemoryRequirement;
	pOSDExtension->DMASafePhysicalAddress.Value =cpu_to_le64(pOSDExtension->dma_handle);

	/* Step5: register resources and interrupt handler */
	if((rc = init_pci_resource(pOSDExtension, pdev))!=0 ){
		goto err_out_iounmap;
	}

	/*Step 6: Assigned resources for components, extension dimm, uncached memory */
	CMPM_AssignComponentExtension(pOSDExtension,(bpVOID)(gShasta_hostp[gNumAdapter]),re_size);
	
	CMPM_AssignComponentResource((bpVOID)(gShasta_hostp[gNumAdapter]));
	
	/*Step 7: Initialize & start components */
	CMPM_InitializeComponents(pOSDExtension,  NULL, pOSDExtension->ControllerInfo.DeviceID, pOSDExtension->ControllerInfo.SubDeviceID);
	
	rc = request_irq(pdev->irq, (void *)shasta_intr, SA_SHIRQ, SHASTA_DEV_NAME, (void *)pOSDExtension);
	if(rc){
	   	printk(KERN_WARNING
			"shasta%x: request_irq()  %d error, rc = %d\n",gNumAdapter, pdev->irq, rc);
	   	goto err_out_iounmap;
   	}

					
	gNumAdapter++;	
	
   	init_timer(&pOSDExtension->InitRAID_Timer);
	return;
err_out_iounmap:
	shasta_iounmap(pOSDExtension);

	pci_free_consistent(pdev,pOSDExtension->DMASafeSize,(void *)pOSDExtension->DMASafeAddress,pOSDExtension->dma_handle);
err_out_vmalloc:
	vfree(gShasta_hostp[gNumAdapter]);
	gShasta_hostp[gNumAdapter] = NULL;
err_out_host:
	scsi_unregister(shpnt);
err_out_regions:
	pci_release_regions(pdev);
err_out:
	pci_disable_device(pdev);

	OSD_DPRINT( ("shasta_detect: return error code %x\n", rc) );

}

static int shasta_detect(Scsi_Host_Template *template) {
	struct pci_dev *pdev = NULL;
	bpOSD_COMP_EXTENSION p;
	U8 cnt;	
	unsigned long flags;
	PREQ_HEAD pReq;
	PREQ_HEAD pReqKM = NULL;
	OSD_TRACE( (" Enter shasta_detect \n") );	

	template->highmem_io = 1;
	for(cnt=0;cnt<MAX_ADAPTER_TYPE;cnt++)
		while ((pdev=pci_find_device(gVenID[cnt], gDevID[cnt],pdev))){
			shasta_ha_init(pdev, template);
		}
	for(cnt=0;cnt<gNumAdapter;cnt++){
		p=CMPM_GetOSDExtensionFromDeviceExtension(gShasta_hostp[cnt]);

		CMPM_StartComponents(p, COMP_TYPE_OSD);
		if(OSDI_ComponentsAreReady(p))
			p->State = SYS_STATE_FULLY_STARTED;
		else{
			free_irq(p->irq, (void *)p);
			shasta_iounmap(p);
			pci_free_consistent(p->pdev,p->DMASafeSize,(void *)p->DMASafeAddress,p->dma_handle);
			scsi_unregister(p->host);
			pci_release_regions(p->pdev);
			pci_disable_device(p->pdev);			
			vfree(gShasta_hostp[cnt]);
			gShasta_hostp[cnt] = NULL;
			}
		}
if(gNumAdapter){
	if ((gShastaMajor = register_chrdev(0, SHASTA_DEV_NAME, (struct file_operations *)&shasta_fops) ) < 0){
		OSD_DPRINT( ("shasta: unable to register \"%s\" character device, error = %d \n", SHASTA_DEV_NAME, gShastaMajor) );

	}
	if(register_reboot_notifier(&shasta_notifier)){
		OSD_DPRINT( ("shasta: unable to register power notifier!\n") );
  	}
#ifdef _64BPLATFORM
	register_ioctl32_conversion(IOCTL_FASTTRAK_GETVERSION,NULL);
	register_ioctl32_conversion(IOCTL_FASTTRAK_PASSTHRU,NULL);
#endif
	}
return gNumAdapter;
}

static int
shasta_release(struct Scsi_Host *host)
{
bpOSD_COMP_EXTENSION p = NULL;

	if(host) p = (bpOSD_COMP_EXTENSION)host->hostdata[0];

	if(p) release_single_device(p);
	scsi_unregister(host);
	gReleaseAdapter ++;
	if(gReleaseAdapter == gNumAdapter){
		if(gShastaMajor > 0)
  			unregister_chrdev(gShastaMajor,SHASTA_DEV_NAME);
		#ifdef _64BPLATFORM
		unregister_ioctl32_conversion(IOCTL_FASTTRAK_GETVERSION);
		unregister_ioctl32_conversion(IOCTL_FASTTRAK_PASSTHRU);
		#endif
		unregister_reboot_notifier(&shasta_notifier);
		}
	return 0;
}

static Scsi_Host_Template driver_template = {
	.module	=THIS_MODULE,
	.name = "shasta",
	.queuecommand = shasta_queue,
	.eh_abort_handler = shasta_abort,
	.eh_device_reset_handler = shasta_reset,
	.eh_host_reset_handler = shasta_reset,
	.bios_param = shasta_biosparam,
	.can_queue = CAN_QUEUE-2,
	.this_id = -1,
	.sg_tablesize = MAX_SG,
	.cmd_per_lun = 1,
	.use_clustering = DISABLE_CLUSTERING/*ENABLE_CLUSTERING*/,	
	.proc_dir = &proc_scsi_shasta,                                               
    	.proc_info = shasta_proc_info, 
	.proc_name = "shasta",
	.detect = shasta_detect,                                        
    	.release = shasta_release,    
    	.use_new_eh_code = 1
};

MODULE_LICENSE("GPL");

#include "scsi_module.c"

#else

/* function   : static int shasta_biosparam
 * input        : Standard Parameters
 * output      : 0
 * description  : SCSI biosparam function
 */
static int shasta_biosparam(struct scsi_device *sdev, struct block_device *bdev, sector_t capacity, int *Geometry)
{
	int heads, sectors; 
	u64 cylinders;
 
	if (capacity > 0x200000) {	
  		heads = 255;	
  		sectors = 63;  	
		cylinders = capacity;  		
   		sector_div(cylinders, (heads*sectors) );  

   	}else{
   		heads = 64;  		
   		sectors = 32;
		cylinders = capacity;  		
   		sector_div(cylinders, (heads*sectors) );  
   	}
	Geometry[0] = heads;
	Geometry[1] = sectors;
	Geometry[2] = (int)cylinders;

	OSD_TRACE( ("capacity = 0x%ullx, Heads = 0x%x, Sectors = 0x%x, Cylinders = 0x%x\n", capacity, heads, sectors, (int)cylinders) );
   	return 0;
}

static struct scsi_host_template driver_template = {
	.module	=THIS_MODULE,
	.name = "shasta",
	.queuecommand = shasta_queue,
	.eh_abort_handler = shasta_abort,
	.eh_device_reset_handler = shasta_reset,
	.eh_host_reset_handler = shasta_reset,
	.bios_param = shasta_biosparam,
	.can_queue = CAN_QUEUE-2,
	.this_id = -1,
	.sg_tablesize = MAX_SG,
	.cmd_per_lun = 1,
	.use_clustering = DISABLE_CLUSTERING/*ENABLE_CLUSTERING*/,
	.proc_name = "shasta"
};

/**
 * shasta_probe - Adapter hot plug add entry point
 *
 * Return value:
 * 	0 on success / non-zero on failure
 **/
static int __devinit shasta_probe(struct pci_dev *pdev,
			       const struct pci_device_id *dev_id)
{
	bpOSD_COMP_EXTENSION pOSDExtension;
	struct Scsi_Host * shpnt;
	u32 rc = PCIBIOS_SUCCESSFUL;
	ULONG DMAMemoryRequirement = 0;
	U32 re_size;
	
	OSD_TRACE( (" Enter shasta_probe \n") );	
	re_size = CMPM_GetComponentExtensionRequirement(gIsHibernation);
	
	/*Step1: enable and initialize pci device*/
	if ((rc = pci_enable_device(pdev))) {
		dev_err(&pdev->dev, "Cannot enable adapter\n");
		return rc;
	}

	rc = pci_request_regions(pdev, SHASTA_DEV_NAME);
	if (rc)
		goto err_out;
	
	rc = set_dma_mask(pdev);
	if (rc)
		goto err_out_regions;

	pci_set_master(pdev);
	
	dev_info(&pdev->dev, "found shasta controller with IRQ: %d\n", pdev->irq);

	/*Step2: alloc and initialize scsi host */

	shpnt = scsi_host_alloc(&driver_template, sizeof(void *));

	if (!shpnt) {
		dev_err(&pdev->dev, "call to scsi_host_alloc failed\n");
		rc = -ENOMEM;
		goto err_out_regions;
	}
	if(st_init_common(&pOSDExtension, shpnt, pdev, re_size)){
		rc = -ENOMEM;
		goto err_out_host;
		}

	DMAMemoryRequirement = 
		CMPM_GetUncachedMemoryRequirement(pOSDExtension, gIsHibernation);
	
	pOSDExtension->DMASafeAddress = pci_alloc_consistent(pdev,DMAMemoryRequirement,&pOSDExtension->dma_handle);
	if(NULL == pOSDExtension->DMASafeAddress){
		OSD_DPRINT( ("Shasta: unable to allocate uncached memory 0x%x\n", DMAMemoryRequirement) );
		rc = -ENOMEM;
		goto err_out_vmalloc;
	}
	pOSDExtension->DMASafeSize = DMAMemoryRequirement;
	pOSDExtension->DMASafePhysicalAddress.Value =cpu_to_le64(pOSDExtension->dma_handle);
	/* Step5: register resources and interrupt handler */
	if((rc = init_pci_resource(pOSDExtension, pdev))!=0 ){
		goto err_out_iounmap;
	}
	
	/*Step 6: Assigned resources for components, extension dimm, uncached memory */
	CMPM_AssignComponentExtension(pOSDExtension,(bpVOID)(gShasta_hostp[gNumAdapter]),re_size);
	
	CMPM_AssignComponentResource((bpVOID)(gShasta_hostp[gNumAdapter]));
		
	/*Step 7: Initialize & start components */
	CMPM_InitializeComponents(pOSDExtension,  NULL, pOSDExtension->ControllerInfo.DeviceID, pOSDExtension->ControllerInfo.SubDeviceID);
	rc = request_irq(pdev->irq, (void *)shasta_intr, SA_SHIRQ, SHASTA_DEV_NAME, (void *)pOSDExtension);
	if(rc){
	   	dev_err(&pdev->dev, "Shasta%x: request_irq()  %d error, rc = %d\n",gNumAdapter, pdev->irq, rc);
	   	goto err_out_iounmap;
   	}

	CMPM_StartComponents(pOSDExtension, COMP_TYPE_OSD);
	if(!OSDI_ComponentsAreReady(pOSDExtension))	{
		rc = -ENXIO;
		goto err_out_freeirq;
		}	
	gNumAdapter++;	
	
   	init_timer(&pOSDExtension->InitRAID_Timer);
#if 0
	disable_irq(pdev->irq);/*don't touch this if you don't know what it means--edl*/
	enable_irq(pdev->irq);/*don't touch this if you don't know what it means--edl*/
#endif 

	rc = scsi_add_host(shpnt, &pdev->dev);
	if (rc)
		goto err_out_freeirq;

	scsi_scan_host(shpnt);
	
	pOSDExtension->State = SYS_STATE_FULLY_STARTED;
	return 0;

err_out_freeirq:
	free_irq(pdev->irq, (void *)pOSDExtension);
err_out_iounmap:
	shasta_iounmap(pOSDExtension);

	pci_free_consistent(pdev,pOSDExtension->DMASafeSize,(void *)pOSDExtension->DMASafeAddress,pOSDExtension->dma_handle);
err_out_vmalloc:
	vfree(gShasta_hostp[gNumAdapter]);
	gShasta_hostp[gNumAdapter] = NULL;
err_out_host:
	scsi_host_put(shpnt);
err_out_regions:
	pci_release_regions(pdev);
err_out:
	pci_disable_device(pdev);

	OSD_DPRINT( ("shasta_probe: return error code %x\n", rc) );

	return rc;
}


/**
 * shasta_remove - adapter hot plug remove entry point
 * @pdev:	pci device struct
 *
 * Adapter hot plug remove entry point.
 *
 * Return value:
 * 	none
 **/
static void __devexit shasta_remove(struct pci_dev *pdev)
{
	bpOSD_COMP_EXTENSION p = (bpOSD_COMP_EXTENSION)pci_get_drvdata(pdev);
	
	scsi_remove_host(p->host);

	release_single_device(p);

}

/**
 * shasta_shutdown - Shutdown handler.
 * @dev:	device struct
 *
 * This function is invoked upon system shutdown/reboot. It will issue
 * an adapter shutdown to the adapter to flush the write cache.
 *
 * Return value:
 * 	none
 **/
static void shasta_shutdown(struct device *dev)
{
	bpOSD_COMP_EXTENSION p = (bpOSD_COMP_EXTENSION)pci_get_drvdata(to_pci_dev(dev) );

	release_single_device(p);
}

#ifdef CONFIG_PM
//static int gPmstate=0;
static int shasta_suspend(struct pci_dev *pdev, u32 state)
{
	bpOSD_COMP_EXTENSION p = (bpOSD_COMP_EXTENSION)pci_get_drvdata(pdev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	 pci_save_state(pdev);
#else
	 pci_save_state(pdev, p->pm_state);
#endif
	 pci_set_power_state (pdev,state);
	return 0;
}

static int shasta_resume(struct pci_dev *pdev)
{

	bpMU_COMP_EXTENSION pMU;
	ULONG DMAMemoryRequirement = 0;
	U32 re_size;
	bpOSD_COMP_EXTENSION p = (bpOSD_COMP_EXTENSION)pci_get_drvdata(pdev);
	//bpOSD_COMP_EXTENSION p1;
  
        pci_set_power_state (pdev,0);
        pci_enable_wake(pdev,0,0);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
	pci_restore_state(pdev);
#else
	pci_restore_state(pdev, p->pm_state);
#endif

	re_size = CMPM_GetComponentExtensionRequirement(gIsHibernation);
        DMAMemoryRequirement=CMPM_GetUncachedMemoryRequirement(p, gIsHibernation);
	p->DMASafeSize = DMAMemoryRequirement;

	if(p->DMASafeAddress) 
		pci_free_consistent(p->pdev,p->DMASafeSize,(void *)p->DMASafeAddress,p->dma_handle);
	p->DMASafeAddress = pci_alloc_consistent(pdev,DMAMemoryRequirement,&p->dma_handle);
	p->DMASafePhysicalAddress.Value =cpu_to_le64(p->dma_handle);

	init_pci_resource(p, pdev);	
		
	CMPM_AssignComponentExtension(p,(p->DeviceExtension),re_size);
	
	CMPM_AssignComponentResource(p->DeviceExtension);
	CMPM_InitializeComponents(p,  NULL, p->ControllerInfo.DeviceID, p->ControllerInfo.SubDeviceID);
	CMPM_RestartComponents(p);
	pMU=CMPM_GetComponentExtension(p,COMP_TYPE_MU);	
	pMU->RequestPayloadHead=0;
	pMU->RequestPayloadTail=0;
	pMU->StatusPayloadHead=0;
	pMU->StatusPayloadTail=0;

        return 0;

}
#endif

static struct pci_device_id shasta_pci_table[] __devinitdata = {
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_SHASTA,
		PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0UL },
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_SHASTA1,
		PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0UL },
	{ PCI_VENDOR_ID_PROMISE, PCI_DEVICE_ID_SHASTA_P,
		PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0UL },
	{0, 0, 0, 0, 0, 0, 0UL},
};

MODULE_AUTHOR("Isaac Huang & Ed Lin");
MODULE_DESCRIPTION("Promise Technology Shasta Controllers");
MODULE_LICENSE("GPL");
MODULE_DEVICE_TABLE(pci, shasta_pci_table);

static struct pci_driver shasta_driver = {
	.name = SHASTA_DEV_NAME,
	.id_table = shasta_pci_table,
	.probe = shasta_probe,
	.remove = shasta_remove,
	.driver = {
		.shutdown = shasta_shutdown,
	},
#ifdef CONFIG_PM
	.suspend	= shasta_suspend,
	.resume	= shasta_resume,
#endif
};

/**
 * shasta_init - Module entry point
 *
 * Return value:
 * 	0 on success / non-zero on failure
 **/
static int __init shasta_init(void)
{
	printk("Promise SuperTrak EX8350/8300 Device Driver %s %s\n", SHASTA_DRIVER_VERSION, SHASTA_DRIVER_DATE);

	if ( (gShastaMajor = register_chrdev(0, SHASTA_DEV_NAME, (struct file_operations *)&shasta_fops) ) < 0){
		OSD_DPRINT( ("shasta: unable to register \"%s\" character device, error = %d \n", SHASTA_DEV_NAME, gShastaMajor) );
		return gShastaMajor;
	}
#ifdef _64BPLATFORM
	register_ioctl32_conversion(IOCTL_FASTTRAK_GETVERSION,NULL);
	register_ioctl32_conversion(IOCTL_FASTTRAK_PASSTHRU,NULL);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,8)
	return pci_module_init(&shasta_driver);
#else
	if (!(pci_register_driver(&shasta_driver)) ) {
        	pci_unregister_driver(&shasta_driver);
        	return -ENODEV;
    	}

	return 0;
#endif
}

/**
 * shasta_exit - Module unload
 *
 * Module unload entry point.
 *
 * Return value:
 * 	none
 **/
static void __exit shasta_exit(void)
{
	if(gShastaMajor > 0)
  		unregister_chrdev(gShastaMajor,SHASTA_DEV_NAME);
#ifdef _64BPLATFORM
	unregister_ioctl32_conversion(IOCTL_FASTTRAK_GETVERSION);
	unregister_ioctl32_conversion(IOCTL_FASTTRAK_PASSTHRU);
#endif
	pci_unregister_driver(&shasta_driver);
}

module_init(shasta_init);
module_exit(shasta_exit);
#endif

