#include "stdafx.h"

void MU_HandShake( bpMU_COMP_EXTENSION pMUExtension );
BOOLEAN MU_PrepareRequestPayload( bpVOID DeviceExtension, PSCSI_REQUEST_BLOCK Srb );
void MU_NotifyRequestHeadChange( bpVOID MUExt, BoOL Interrupt );
BOOLEAN MU_HandleStatusPayload( bpVOID DeviceExtension, bpVOID Context );
void MU_TriggerInterrupt( bpMU_COMP_EXTENSION pMUExtension, U32 Interrupt );

bpMU_COMP_EXTENSION gMUExtension = NULL;

#ifdef SUPPORT_DEBUG_FRAME
volatile PDEBUG_FRAME pDebugFrame = NULL;
#endif

extern void Host_OSD_Delay(U32 DelayTime);
extern void OSD_GetCurrentTime(PU64 time);

#if defined(_OS_WINDOWS)
	#define HMU_PARTNER_TYPE	HMU_PARTNER_TYPE_WINDOWS
#elif defined(_OS_LINUX)
	#define HMU_PARTNER_TYPE	HMU_PARTNER_TYPE_LINUX
#elif defined(_OS_BIOS)
	#define HMU_PARTNER_TYPE	HMU_PARTNER_TYPE_BIOS
#elif defined(_OS_DOS)
	#define HMU_PARTNER_TYPE	HMU_PARTNER_TYPE_DOS
#endif

/* 
 * Rotate the request/status payload head/tail pointers.
 */
#define MU_ROTATE_INCREASE_POINTER(pCurrentValue)	do {	\
		(*(pCurrentValue))++;								\
		if ( Host_OSD_IsLdrDump(NULL) )						\
		{													\
			if ( *(pCurrentValue)>=MU_MIN_PAYLOAD_COUNT )	\
				*(pCurrentValue)=0;							\
		}else												\
		{													\
			if ( *(pCurrentValue)>=MU_MAX_PAYLOAD_COUNT )	\
				*(pCurrentValue)=0;							\
		}													\
	}while(0)



/************************************************************
 *				Message Unit Component Interface			*
 ************************************************************/
void 
MU_ComponentGetExtensionSize(
	OUT bpMEMORY_REQUIREMENT pMemRequirement 
	)
{
	pMemRequirement->MaxRequirement = sizeof(MU_COMP_EXTENSION);
	pMemRequirement->MinRequirement = sizeof(MU_COMP_EXTENSION);
}

#define MU_MAX_REQUEST_BUFFER_SIZE	( (MU_MAX_PAYLOAD_COUNT) * sizeof(MU_REQUEST_PAYLOAD) )
#define MU_MIN_REQUEST_BUFFER_SIZE	( (MU_MIN_PAYLOAD_COUNT) * sizeof(MU_REQUEST_PAYLOAD) )
#define MU_MAX_STATUS_BUFFER_SIZE	( (MU_MAX_PAYLOAD_COUNT) * sizeof(MU_STATUS_PAYLOAD) )
#define MU_MIN_STATUS_BUFFER_SIZE	( (MU_MIN_PAYLOAD_COUNT) * sizeof(MU_STATUS_PAYLOAD) )

void
MU_ComponentGetDMASafeMemoryRequirement(
	IN U32 DeviceID,
	IN U32 SubDeviceID,
	OUT bpMEMORY_REQUIREMENT pMemRequirement 
	)
{
	pMemRequirement->MaxRequirement = 
		MU_MAX_REQUEST_BUFFER_SIZE + MU_MAX_STATUS_BUFFER_SIZE
		+8;	// For DWORD Alignment.

	pMemRequirement->MinRequirement = 
		MU_MIN_REQUEST_BUFFER_SIZE + MU_MIN_STATUS_BUFFER_SIZE
		+8;	// For DWORD Alignment.

	//OSD_TRACE(("Request Payload Size = 0x%x.\n", sizeof(MU_REQUEST_PAYLOAD)));
	//OSD_TRACE(("Status Payload Size = 0x%x.\n", sizeof(MU_STATUS_PAYLOAD)));

	//OSD_TRACE(("MU_MAX_PAYLOAD_COUNT = 0x%x.\n", MU_MAX_PAYLOAD_COUNT));
	//OSD_TRACE(("MU_MAX_REQUEST_BUFFER_SIZE = 0x%x.\n", MU_MAX_REQUEST_BUFFER_SIZE));
	//OSD_TRACE(("MU_MAX_STATUS_BUFFER_SIZE = 0x%x.\n", MU_MAX_STATUS_BUFFER_SIZE));

	//OSD_TRACE(("MU_MIN_PAYLOAD_COUNT = 0x%x.\n", MU_MIN_PAYLOAD_COUNT));
	//OSD_TRACE(("MU_MIN_REQUEST_BUFFER_SIZE = 0x%x.\n", MU_MIN_REQUEST_BUFFER_SIZE));
	//OSD_TRACE(("MU_MIN_STATUS_BUFFER_SIZE = 0x%x.\n", MU_MIN_STATUS_BUFFER_SIZE));
}

void
MU_ComponentInitialize(
	IN bpVOID MUExtension,
	IN bpMEMORY_AVAILABLE pMemAvail,
	IN U32 DeviceID,
	IN U32 SubDeviceID,
	IN bpVOID ComponentSpecificContext
)
{
	/* Initialize the MU Extension */
	bpMU_COMP_EXTENSION pMUExtension = (bpMU_COMP_EXTENSION)MUExtension;
	bpMU_INITIALIZE_CONTEXT bpBaseAddr = (bpMU_INITIALIZE_CONTEXT)ComponentSpecificContext;
	U16 i;
	U32 temp;

	gMUExtension = pMUExtension;

	OSD_DASSERT( pMemAvail->ComponentExtensionSize>=sizeof(MU_COMP_EXTENSION) );

	pMUExtension->pOSDExtension = 
		(bpOSD_COMP_EXTENSION)CMPM_GetComponentExtension(pMUExtension, COMP_TYPE_OSD);

	/* ATU Register Base Address */
	pMUExtension->BaseAddress = bpBaseAddr->BaseAddress[0];
//#ifndef CONSTRAIN_SDRAM_BASE_ADDRESS
	/* We use PCI Base Address 0 and 2. 
	 * But since 1 is empty, windows move 2 to 1. */
	pMUExtension->SDRAMBase = bpBaseAddr->BaseAddress[1];
//#endif

	/* QWORD Align. */
	temp = (U32)(ALIGN_ROUND((PTR_SIZE)pMemAvail->DMASafeAddress, 8) - (PTR_SIZE)pMemAvail->DMASafeAddress);
	pMemAvail->DMASafeAddress = (bpVOID)((PTR_SIZE)pMemAvail->DMASafeAddress + temp);
    pMemAvail->DMASafePhysicalAddress = U64_Add_U32(pMemAvail->DMASafePhysicalAddress, temp);
	OSD_DASSERT((PTR_SIZE)pMemAvail->DMASafePhysicalAddress.u.Low == ALIGN_ROUND((PTR_SIZE)pMemAvail->DMASafePhysicalAddress.u.Low, 8));

	/* Request Payload Window and Status Payload Window */
	pMUExtension->RequestBufferAddr = pMemAvail->DMASafeAddress;
	pMUExtension->RequestBufferPhyAdd = pMemAvail->DMASafePhysicalAddress;

	if ( !Host_OSD_IsLdrDump(NULL) )
	{
		pMUExtension->RequestPayloadCount = MU_MAX_PAYLOAD_COUNT;
		pMUExtension->StatusPayloadCount = MU_MAX_PAYLOAD_COUNT;

		pMUExtension->StatusBufferAddr = (PU8)pMemAvail->DMASafeAddress
			+ MU_MAX_REQUEST_BUFFER_SIZE;
		OSD_DASSERT(pMemAvail->DMASafePhysicalAddress.u.High==0);
		pMUExtension->StatusBufferPhyAddr.u.High = 0;
		pMUExtension->StatusBufferPhyAddr.u.Low = 
			pMemAvail->DMASafePhysicalAddress.u.Low
			+ MU_MAX_REQUEST_BUFFER_SIZE;

		OSD_DASSERT( (MU_MAX_REQUEST_BUFFER_SIZE+MU_MAX_STATUS_BUFFER_SIZE)
			<=pMemAvail->DMASafeSize );
	}
	else
	{
		/* We'll waste on payload. */
		pMUExtension->RequestPayloadCount = MU_MIN_PAYLOAD_COUNT;
		pMUExtension->StatusPayloadCount = MU_MIN_PAYLOAD_COUNT;

		pMUExtension->StatusBufferAddr = (PU8)pMemAvail->DMASafeAddress
			+ MU_MIN_REQUEST_BUFFER_SIZE;
		OSD_DASSERT(pMemAvail->DMASafePhysicalAddress.u.High==0);
		pMUExtension->StatusBufferPhyAddr.u.High = 0;
		pMUExtension->StatusBufferPhyAddr.u.Low = 
			pMemAvail->DMASafePhysicalAddress.u.Low
			+ MU_MIN_REQUEST_BUFFER_SIZE;

		OSD_DASSERT( (MU_MIN_REQUEST_BUFFER_SIZE+MU_MIN_STATUS_BUFFER_SIZE)
			<=pMemAvail->DMASafeSize );
	}

	Tag_Initialize(&pMUExtension->CmdTag);
	pMUExtension->OutstandingRequestCount = 0;
	pMUExtension->BusySRBCount = 0;
	pMUExtension->ComponentState = MU_STATE_STARTING;

	for ( i=0; i<MU_MAX_REQUEST; i++ )
		pMUExtension->OutstandingRequest[i] = NULL;
}

void
MU_ComponentStop(
	IN bpVOID MUExtension
)
{
	bpMU_COMP_EXTENSION pMUExtension = (bpMU_COMP_EXTENSION)MUExtension;
	U32 i;

	/* sanity check */
	for ( i=0; i<MU_MAX_REQUEST; i++ )
	{
        OSD_DASSERT( pMUExtension->OutstandingRequest[i]==NULL );
	}
	OSD_DASSERT( pMUExtension->CmdTag==0 );
	OSD_DASSERT( pMUExtension->OutstandingRequestCount==0 );

	/* reset the data structure */
	pMUExtension->ComponentState = MU_STATE_STARTING;
	pMUExtension->BusySRBCount = 0;

	pMUExtension->RequestPayloadHead = 0;
	pMUExtension->RequestPayloadTail = 0;
	pMUExtension->StatusPayloadHead = 0;
	pMUExtension->StatusPayloadTail = 0;

#if 0
	/* Sign Firmware We've been stopped. */
	MU_TriggerInterrupt(pMUExtension, MU_INBOUND_DOORBELL_HMUSTOPPED);	
#endif
}

void
MU_ComponentStart(
	IN bpVOID MUExtension
)
{
	bpMU_COMP_EXTENSION pMUExtension = (bpMU_COMP_EXTENSION)MUExtension;
	U32 i;

	/* sanity check */
	for ( i=0; i<MU_MAX_REQUEST; i++ )
	{
		OSD_DASSERT( pMUExtension->OutstandingRequest[i]==NULL );
	}
	OSD_DASSERT( pMUExtension->CmdTag==0 );
	OSD_DASSERT( pMUExtension->OutstandingRequestCount==0 );

#ifdef FINISH_HANDSHAKE_FIRST
#ifndef USE_STORPORT
	/* Wait 3 seconds for handshake to finish. */
	#define MU_MAX_DELAY_TIME 3000
#else
	/* STORPORT needs special handling if complete request with SRB_STATUS_BUSY. */
	#define MU_MAX_DELAY_TIME 15000
#endif

	for ( i=0; i<MU_MAX_DELAY_TIME; i++ )
	{
		MU_HandShake(pMUExtension);

		if ( pMUExtension->ComponentState==MU_STATE_STARTED )
		{
			OSD_DPRINT(("Handshake is finished after waiting for %d milli-second.\n", i));
			break;
		}

		Host_OSD_Delay(1000);
	}

	//OSD_DASSERT( pMUExtension->ComponentState==MU_STATE_STARTED );
#else
	/* MU notifies OSD it has been started before handshake. 
	 * So it won't block the system bootup. */
	MU_HandShake(pMUExtension);
#endif

#if defined(_LINUXDRIVER)
	if ( pMUExtension->ComponentState==MU_STATE_STARTED )
#endif
	OSD_ComponentStarted(pMUExtension, COMP_START_STATUS_SUCCESS);
}

void 
MU_ComponentRestart(
	IN bpVOID MUExtension
)
{
	/* Ask firmware to do reset. */
	bpMU_COMP_EXTENSION pMUExtension = (bpMU_COMP_EXTENSION)MUExtension;
	MU_TriggerInterrupt(pMUExtension, MU_INBOUND_DOORBELL_RESET);	

	/* It's just like MU_ComponentStart. */
	MU_ComponentStart( MUExtension );
}

/************************************************************
 *				Message Unit Startup Functions				*
 ************************************************************/
VOID 
MU_TimerService(
	IN bpVOID HwDeviceExtension
	)
{
	//bpOSD_COMP_EXTENSION pOSDExtension = (bpOSD_COMP_EXTENSION)
	//	CMPM_GetOSDExtensionFromDeviceExtension(HwDeviceExtension);
	//OSD_RequestTimer(pOSDExtension, MU_TimerService, 5000000);
	OSD_TRACE(("MU_TimerService.\n"));
}

#ifdef SUPPORT_DEBUG_FRAME
void
MU_GetDebugFrame(
	bpMU_COMP_EXTENSION pMUExtension
	)
{
	PTR_SIZE BaseAddress = (PTR_SIZE)pMUExtension->BaseAddress;
	PTR_SIZE SDRAMBase = (PTR_SIZE)pMUExtension->SDRAMBase;
	PTR_SIZE FrameAddress = OSD_HwReadU32(BaseAddress+MU_OUTBOUND_MESSAGE_REG1);

#if 0
	OSD_DASSERT( FrameAddress!=0 );

	pDebugFrame = (PDEBUG_FRAME)( FrameAddress+(PTR_SIZE)pMUExtension->SDRAMBase ) ; 
	
	/* FMU */
	OSD_DASSERT( pDebugFrame->fmu_local_payload!=NULL
		&& pDebugFrame->fmu_state!=NULL
		&& pDebugFrame->fmu_request_head!=NULL
		&& pDebugFrame->fmu_request_tail!=NULL );

	pDebugFrame->fmu_local_payload = (bpMSG *)((PTR_SIZE)pDebugFrame->fmu_local_payload + SDRAMBase);
	pDebugFrame->fmu_state = (PU32)((PTR_SIZE)pDebugFrame->fmu_state + SDRAMBase);
	pDebugFrame->fmu_request_head = (PU32)((PTR_SIZE)pDebugFrame->fmu_request_head + SDRAMBase);
	pDebugFrame->fmu_request_tail = (PU32)((PTR_SIZE)pDebugFrame->fmu_request_tail + SDRAMBase);

	/* Kernel */
	pDebugFrame->g_free_msg_pool = (bpGenPool)((PTR_SIZE)pDebugFrame->g_free_msg_pool + SDRAMBase);
	pDebugFrame->g_free_req_pool = (bpGenPool)((PTR_SIZE)pDebugFrame->g_free_req_pool + SDRAMBase);
	pDebugFrame->g_inbound_msg_list = (bpGEN_LIST_HEADER)((PTR_SIZE)pDebugFrame->g_inbound_msg_list + SDRAMBase);
	pDebugFrame->g_outbound_msg_list = (bpGEN_LIST_HEADER)((PTR_SIZE)pDebugFrame->g_outbound_msg_list + SDRAMBase);
	pDebugFrame->g_complete_task_list = (bpGEN_LIST_HEADER)((PTR_SIZE)pDebugFrame->g_complete_task_list + SDRAMBase);
	pDebugFrame->g_cam_task_count = (PU32)((PTR_SIZE)pDebugFrame->g_cam_task_count + SDRAMBase);

	/* CAM */
	pDebugFrame->g_CAM = (bpCAM)((PTR_SIZE)pDebugFrame->g_CAM + SDRAMBase);
	pDebugFrame->g_XorPendingList = (bpGEN_LIST_HEADER)((PTR_SIZE)pDebugFrame->g_XorPendingList + SDRAMBase);
	pDebugFrame->g_XorPingList = (bpGEN_LIST_HEADER)((PTR_SIZE)pDebugFrame->g_XorPingList + SDRAMBase);
	pDebugFrame->g_XorPongList = (bpGEN_LIST_HEADER)((PTR_SIZE)pDebugFrame->g_XorPongList + SDRAMBase);
	pDebugFrame->g_HdmaPendingList = (bpGEN_LIST_HEADER)((PTR_SIZE)pDebugFrame->g_HdmaPendingList + SDRAMBase);
	pDebugFrame->g_HdmaPingList = (bpGEN_LIST_HEADER)((PTR_SIZE)pDebugFrame->g_HdmaPingList + SDRAMBase);

	/* OSD */
	pDebugFrame->g_OSD = (bpOSD_COMP_EXTENSION)((PTR_SIZE)pDebugFrame->g_OSD + SDRAMBase);

	/* Engine */
	pDebugFrame->g_Engine = (bpRaidEngine)((PTR_SIZE)pDebugFrame->g_Engine + SDRAMBase);

	/* PBM */
	pDebugFrame->g_CC = (CC_DEF*)((PTR_SIZE)pDebugFrame->g_CC + SDRAMBase);
#endif
}
#endif

void
MU_HandShake(
	bpMU_COMP_EXTENSION pMUExtension
	)
{
	U32 Data;
	bpHMU_HANDSHAKE_FRAME bpHMUFrame;
	//bpFMU_HANDSHAKE_FRAME bpFMUFrame;
	PTR_SIZE BaseAddress = (PTR_SIZE)pMUExtension->BaseAddress;
	//OSD_DPRINT(("MU_HandShake ComponentState = 0x%x.\n", pMUExtension->ComponentState));
	switch ( pMUExtension->ComponentState ) 
	{
		case MU_STATE_STARTING:
			/* Read the OMR0 to check whether the FMU is running. */
			Data = OSD_HwReadU32(BaseAddress + MU_OUTBOUND_MESSAGE_REG0);
			//OSD_DPRINT(("Handshake: Outbound Message Register 0: 0x%x.\n", Data));
			if ( Data!=MU_HANDSHAKE_SIGNATURE )
			{
				/* For multiple handshake, trigger FMU to handshake again. */
				MU_TriggerInterrupt(pMUExtension, MU_INBOUND_DOORBELL_HANDSHAKE);
				return;
			}

			pMUExtension->ComponentState = MU_STATE_FMU_READY_FOR_HANDSHAKE;
			/* Intentionally there is no break. */

		case MU_STATE_FMU_READY_FOR_HANDSHAKE:
			/* Send the handshake frame to the FMU */
			bpHMUFrame = (bpHMU_HANDSHAKE_FRAME)pMUExtension->StatusBufferAddr;

			bpHMUFrame->RequestBufferPhyAddr = pMUExtension->RequestBufferPhyAdd.u.Low;
			bpHMUFrame->RequestBufferPhyAddrHigh = pMUExtension->RequestBufferPhyAdd.u.High;
			bpHMUFrame->RequestPayloadSize = sizeof(MU_REQUEST_PAYLOAD);
			bpHMUFrame->RequestPayloadCount = pMUExtension->RequestPayloadCount;
			bpHMUFrame->StatusPayloadSize = sizeof(MU_STATUS_PAYLOAD);
			bpHMUFrame->StatusPayloadCount = pMUExtension->StatusPayloadCount;
			OSD_GetCurrentTime(&bpHMUFrame->HostTime);
			#if defined(_OS_WINDOWS)
				if ( !gIsHibernation )
					bpHMUFrame->PartnerType = HMU_PARTNER_TYPE_WINDOWS;
				else
					bpHMUFrame->PartnerType = HMU_PARTNER_TYPE_WINDOWS_HIBERNATION;
			#else
            	bpHMUFrame->PartnerType = HMU_PARTNER_TYPE;
			#endif
			
			/* Notify FMU the Base Address of the Handshake Frame */
			OSD_HwWriteU32(BaseAddress + MU_INBOUND_MESSAGE_REG0,
				pMUExtension->StatusBufferPhyAddr.u.Low);
			OSD_HwWriteU32(BaseAddress + MU_OUTBOUND_MESSAGE_REG0,
				pMUExtension->StatusBufferPhyAddr.u.High);

			/* Signal a interrupt to the FMU */
			MU_TriggerInterrupt(pMUExtension, MU_INBOUND_DOORBELL_HANDSHAKE);

			pMUExtension->ComponentState = MU_STATE_SEND_HANDSHAKE_FRAME;
            return;

		case MU_STATE_SEND_HANDSHAKE_FRAME:
			/* Read the OMR0 to check whether the FMU has got the Handshake Frame. */
			Data = OSD_HwReadU32(BaseAddress + MU_OUTBOUND_MESSAGE_REG0);
			//OSD_DPRINT(("Handshake: Outbound Message Register 0: 0x%x.\n", Data));

			/* It's not triggered by interrupt. It's invoked actively by Driver itself. */
			if ( Data!=MU_HANDSHAKE_SIGNATURE )
				return;

			break;
		#ifdef _LINUXDRIVER
		case MU_STATE_STARTED:
			return;/*maybe a bug of compiler? smp rush? need to lock in init? investigate later--edl*/
		#endif
		default:
			OSD_DASSERT(FALSE);
	}

	/* Handshake is completed. */
	pMUExtension->ComponentState = MU_STATE_STARTED;

	#ifdef SUPPORT_DEBUG_FRAME
	MU_GetDebugFrame(pMUExtension);
	#endif

	/* Initialize the Request and Status Data Structure */
	OSD_HwWriteU32(BaseAddress+MU_INBOUND_MESSAGE_REG0, 0);
	OSD_HwWriteU32(BaseAddress+MU_OUTBOUND_MESSAGE_REG0, 0);
	OSD_HwWriteU32(BaseAddress+MU_INBOUND_MESSAGE_REG1, 0);
	OSD_HwWriteU32(BaseAddress+MU_OUTBOUND_MESSAGE_REG1, 0);
	/* Ask Windows to re-scan the bus. */
	OSD_NotifyBusChange(pMUExtension);
	OSD_DPRINT(("Handshake is finished.\n"));
	return;
}

/************************************************************
 *			Message Unit Handle Request Functions			*
 ************************************************************/
#define WAIT_HANDSHAKE_MAX_SRB_COUNT	100	/* About 1-2 minutes*/

NEEDCOMPLETE
MU_HandleRequest(
	IN PVOID MUExt,
	IN PSCSI_REQUEST_BLOCK Srb
	)
{
	bpMU_COMP_EXTENSION pMUExtension = (bpMU_COMP_EXTENSION)MUExt;

	if ( pMUExtension->ComponentState!=MU_STATE_STARTED )
		MU_HandShake(pMUExtension);

	/* Handshake is not done yet. We respond the request on behalf of FMU. 
	 * At this moment, we need handle it carefully.
	 * We cannot immediately return SRB_STATUS_SELECTION_TIMEOUT or SRB_STATUS_INVALID_REQUEST.
	 * Otherwise ACPI may fail because it cannot take this kind of response. So we return SRB_STATUS_BUSY.
	 * But we cannot return SRB_STATUS_BUSY forever. 
	 * Otherwise if firmware is dead, windows cannot bootup and retry our driver forever. */
	if ( pMUExtension->ComponentState!=MU_STATE_STARTED )
	{
		if ( pMUExtension->BusySRBCount<WAIT_HANDSHAKE_MAX_SRB_COUNT )
		{
			pMUExtension->BusySRBCount++;
			Srb->SrbStatus=SRB_STATUS_BUSY;
		}
		else 
		{
			if ( Srb->Cdb[0]==scsiINQUIRY ) /* Inquiry Request */
				Srb->SrbStatus=SRB_STATUS_SELECTION_TIMEOUT;
			else
				Srb->SrbStatus=SRB_STATUS_INVALID_REQUEST;
		}

		return TRUE;
	}

	/* Handshake is done. So we submit requests to the FMU. */
	MU_SendRequest(pMUExtension, Srb);
	/* We should complete this request in the Interrupt Service Routine. */
	return FALSE;
}

/* Send this Request to the FMU. */
BoOL
MU_SendRequest(
	IN PVOID MUExt,
	IN PSCSI_REQUEST_BLOCK Srb
)
{
	bpMU_COMP_EXTENSION pMUExtension = (bpMU_COMP_EXTENSION)MUExt;
	bpOSD_COMP_EXTENSION pOSDExtension = (bpOSD_COMP_EXTENSION)
		CMPM_GetComponentExtension(pMUExtension, COMP_TYPE_OSD);

	OSD_DASSERT(pMUExtension->ComponentState==MU_STATE_STARTED);

	/* Prepare a Request Payload */
#if defined(USE_STORPORT) && defined(FULLDUPLEX_STORPORT)
	StorPortSynchronizeAccess(
		pOSDExtension->DeviceExtension,
		(PSTOR_SYNCHRONIZED_ACCESS)MU_PrepareRequestPayload,
		Srb);
	#if VDEBUG
	{
		/* I just want to see whether Timer works in STORPORT. 
		 * Conclusion: Timer works pretty well under STORPORT. */
		static U8 needTimer = FALSE;
		if ( !needTimer )
			OSD_RequestTimer(pOSDExtension, MU_TimerService, 5000000);
		needTimer = TRUE;
	}
	#endif
#else
	MU_PrepareRequestPayload(pOSDExtension->DeviceExtension, Srb);
#endif

	/* Trigger an interrupt to the FMU. */
	MU_NotifyRequestHeadChange(pMUExtension, TRUE);

	return TRUE;
}

bpMU_REQUEST_PAYLOAD
MU_AllocateRequestPayload(
	bpMU_COMP_EXTENSION pMUExtension
	)
{
	/* Suppose we should have enough Request Payload Buffer to hold all the requests. */
	bpMU_REQUEST_PAYLOAD pRequest = 
		((bpMU_REQUEST_PAYLOAD)pMUExtension->RequestBufferAddr)+
		pMUExtension->RequestPayloadHead;
	MU_ROTATE_INCREASE_POINTER(&pMUExtension->RequestPayloadHead);

	#ifndef NO_TAIL_INTERRUPT
	/* Agreement with FMU: RequestPayloadHead=RequestPayloadTail means no request. 
	 * RequestPayloadHead is the position for the next request payload. */
	OSD_DASSERT(pMUExtension->RequestPayloadHead!=pMUExtension->RequestPayloadTail);

	OSD_TRACE((
		"MU_AllocateRequestPayload now RequestPayloadHead=0x%x RequestPayloadTail=0x%x.\n", 
		pMUExtension->RequestPayloadHead,
		pMUExtension->RequestPayloadTail
		));
	#endif

	return pRequest;
}

bpMU_STATUS_PAYLOAD
MU_GetStatusPayload(
	bpMU_COMP_EXTENSION pMUExtension,
	U32 pointer
	)
{
	OSD_DASSERT(pMUExtension->StatusPayloadTail!=pMUExtension->StatusPayloadHead);

	bpMU_STATUS_PAYLOAD pStatus = 
		((bpMU_STATUS_PAYLOAD)pMUExtension->StatusBufferAddr)+pointer;

	MU_ROTATE_INCREASE_POINTER(&pMUExtension->StatusPayloadTail);

	#if 0
	OSD_TRACE((
		"MU_GetStatusPayload Address=0x%x, buffer is from 0x%x.\n",
		(PTR_SIZE)pStatus,
		pMUExtension->StatusBufferAddr
		));
	#endif

	return pStatus;
}

BOOLEAN
MU_PrepareRequestPayload(
	IN PVOID DeviceExtension,
	IN PSCSI_REQUEST_BLOCK Srb
)
{
	bpOSD_COMP_EXTENSION pOSDExtension = (bpOSD_COMP_EXTENSION)
		CMPM_GetOSDExtensionFromDeviceExtension(DeviceExtension);
	bpMU_COMP_EXTENSION pMUExtension = (bpMU_COMP_EXTENSION)
		CMPM_GetComponentExtension(pOSDExtension, COMP_TYPE_MU);
	bpMU_REQUEST_PAYLOAD pRequest
		= MU_AllocateRequestPayload(pMUExtension);

	/* Payload Header */
	pRequest->TaskTag = (U16)Tag_Allocate(&pMUExtension->CmdTag);
	pRequest->LUN = Srb->Lun;
	pRequest->Target = Srb->TargetId;
	pRequest->TaskAttribute = TASK_ATTRIBUTE_SIMPLE;
	pRequest->TaskManagement = 0; /* Not supported yet. */
	pRequest->PayloadSize = (U8)(sizeof(MU_REQUEST_PAYLOAD))/sizeof(U32);	/* Currently always set to the max size. */

	/* CDB */
	gen_memmove(pRequest->CDB, Srb->Cdb, MU_CDB_LENGTH);

	if ( Srb->DataTransferLength!=0 )	/* Flash Request: Buffer is not from the Srb. */
		OSD_PrepareSG( pOSDExtension, Srb, (bpSGTABLE)&pRequest->VariableData );

	/* Our API Payload will be transferred to SG table as well. */
	#if VDEBUG
	//if ( Srb->TargetId==CONSOLEDEVICE_TARGETID )
	//{
		OSD_DumpSrb(Srb);
	//}
	#endif
	pMUExtension->OutstandingRequest[pRequest->TaskTag] = Srb;

	pMUExtension->OutstandingRequestCount++;

	return TRUE;
}

/*
 * MU_INBOUND_MESSAGE_REG0 = Request Payload Head
 * MU_INBOUND_MESSAGE_REG1 = Status Payload Tail
 * MU_OUTBOUND_MESSAGE_REG0 = Request Payload Tail
 * MU_OUTBOUND_MESSAGE_REG1 = Status Payload Head
 * These Register won't trigger the interrupt. I'll mask the interrupt.
 * I'm using the Inbound and Outbound Doorbell Register to sign Interrupt.
 */
void
MU_NotifyRequestHeadChange(
	IN PVOID MUExt,
	IN BoOL Interrupt	/* Besides updating the MU_INBOUND_MESSAGE_REG0, do you want to interrupt FMU. */
)
{
	bpMU_COMP_EXTENSION pMUExtension = (bpMU_COMP_EXTENSION)MUExt;
	PTR_SIZE BaseAddress = (PTR_SIZE)pMUExtension->BaseAddress;
#if 0
	U32 waitCount = 100000;

	while( (OSD_HwReadU32(BaseAddress + MU_INBOUND_DOORBELL) & MU_INBOUND_DOORBELL_REQUESTHEADCHANGED)
		&& waitCount-- )
	{
		ScsiPortStallExecution(1);	// wait FMU to handle the interruot
	}
#endif

	#if 0
	U32 OriginalRequestHead = OSD_HwReadU32(BaseAddress + MU_INBOUND_MESSAGE_REG0);
	OSD_DASSERT( OriginalRequestHead!=pMUExtension->RequestPayloadHead );
	#endif

	/* Update the Inbound Message Register */
	OSD_HwWriteU32(
		BaseAddress + MU_INBOUND_MESSAGE_REG0,
		pMUExtension->RequestPayloadHead 
		);

	/* Trigger Interrupt */
	if ( Interrupt ) 
	{
		//OSD_TRACE(("Signal FMU there are requests waiting for handling.\n"));
		MU_TriggerInterrupt(pMUExtension, MU_INBOUND_DOORBELL_REQUESTHEADCHANGED);
	}
}

void
MU_NotifyStatusTailChange(
	IN PVOID MUExt
)
{
	bpMU_COMP_EXTENSION pMUExtension = (bpMU_COMP_EXTENSION)MUExt;
	PTR_SIZE BaseAddress = (PTR_SIZE)pMUExtension->BaseAddress;

	/* Tell FMU Status Payload Tail is changed. */
	OSD_HwWriteU32(
		BaseAddress + MU_INBOUND_MESSAGE_REG1,
		pMUExtension->StatusPayloadTail
		);

	#ifndef NO_TAIL_INTERRUPT
	MU_TriggerInterrupt(pMUExtension, MU_INBOUND_DOORBELL_STATUSTAILCHANGED);
	#endif
}

/************************************************************
 *			Message Unit Interrupt Service Routine			*
 ************************************************************/
void MU_HandleInterrupt(
	IN bpVOID MUExt,
	U32 Doorbell
)
{
	bpMU_COMP_EXTENSION pMUExtension = (bpMU_COMP_EXTENSION)MUExt;
	bpOSD_COMP_EXTENSION pOSDExtension = (bpOSD_COMP_EXTENSION)
		CMPM_GetComponentExtension(pMUExtension, COMP_TYPE_OSD);
	PTR_SIZE BaseAddress = (PTR_SIZE)pMUExtension->BaseAddress;

	if ( pMUExtension->ComponentState!=MU_STATE_STARTED )
	{
		MU_HandShake(pMUExtension);
		return;
	}

	/* When shut down, we got incorrect interrupt. Ignore this interrupt. */
	if ( pMUExtension->OutstandingRequestCount==0 )
		return;

	/* It's better to return request first before check the Request Tail Pointer */
	if ( Doorbell&MU_OUTBOUND_DOORBELL_STATUSHEADCHANGED )
	{
		/* We got some Status Payloads */
		pMUExtension->StatusPayloadHead = 
			OSD_HwReadU32( BaseAddress+MU_OUTBOUND_MESSAGE_REG1);

		#if 0
		OSD_TRACE((
			"StatusPayloadHead=0x%x StatusPayloadTail=0x%x.\n", 
			pMUExtension->StatusPayloadHead,
			pMUExtension->StatusPayloadTail
			));
		#endif

		#if 0 /* defined(USE_STORPORT) && defined(FULLDUPLEX_STORPORT) */
		/* I suppose ISR under fullduplex is running at DIRQL.
		 * Needn't synchronize it with StartIO which is at DISPATCH level. 
		 * And ISR won't be interrupted by ISR because it holds the interrupt spinlock. */
		StorPortSynchronizeAccess(
					pOSDExtension->DeviceExtension,
					(PSTOR_SYNCHRONIZED_ACCESS)MU_HandleStatusPayload,
					NULL);
		#else
		MU_HandleStatusPayload(pOSDExtension->DeviceExtension, NULL);
		#endif
	}

	if ( Doorbell&MU_OUTBOUND_DOORBELL_REQUESTTAILCHANGED )
	{
		#ifdef NO_TAIL_INTERRUPT
			OSD_DASSERT(FALSE);
		#else
			/* NO_TAIL_INTERRUPT is not defined. */

			/* Update the Request Tail Pointer */
			pMUExtension->RequestPayloadTail = 
				OSD_HwReadU32( BaseAddress+MU_OUTBOUND_MESSAGE_REG0);
			#if 0
			OSD_TRACE((
				"RequestPayloadHead=0x%x, RequestPayloadTail=0x%x.\n", 
				pMUExtension->RequestPayloadHead,
				pMUExtension->RequestPayloadTail
				));
			#endif

		#endif
	}

	return;
}

/* Let's Handle the status payload and return OS Request. */
BOOLEAN
MU_HandleStatusPayload(
	IN bpVOID DeviceExtension,
	IN bpVOID Context
)
{
	bpOSD_COMP_EXTENSION pOSDExtension = (bpOSD_COMP_EXTENSION)
		CMPM_GetOSDExtensionFromDeviceExtension(DeviceExtension);
	bpMU_COMP_EXTENSION pMUExtension = (bpMU_COMP_EXTENSION)
		CMPM_GetComponentExtension(pOSDExtension, COMP_TYPE_MU);
	bpMU_STATUS_PAYLOAD pStatus = NULL;
	PSCSI_REQUEST_BLOCK Srb = NULL;
	PTR_SIZE BaseAddress = (PTR_SIZE)pMUExtension->BaseAddress;
	U32 VariableDataSize = 0;

	while ( pMUExtension->StatusPayloadTail!=pMUExtension->StatusPayloadHead )
	{
		/* The StatusPayloadTail is pointed to the Status Payload 
		 * MU_GetStatusPayload will move the StatusPayloadTail. */
		pStatus = MU_GetStatusPayload(pMUExtension, pMUExtension->StatusPayloadTail);

		OSD_DASSERT( pStatus->TaskTag<TAG_BITMAP_LENGTH );
		OSD_DASSERT( Tag_InUse(&pMUExtension->CmdTag, pStatus->TaskTag) );

		Srb = (PSCSI_REQUEST_BLOCK)pMUExtension->OutstandingRequest[pStatus->TaskTag];
		OSD_DASSERT( Srb!=NULL );

		Srb->SrbStatus = pStatus->SrbStatus;
		Srb->ScsiStatus = pStatus->ScsiStatus;

		/* 
		 * Base on the PayloadSize to judge whether we need copy data here. 
		 * All the commands with DATA will use SG Table.
		 * The Status Payload Variable Data is used for the following purposes only.
		 * 1. Sensor Data. I'll copy the data to the Srb->SenseInfoBuffer.
		 * 2. Some Data for SCSIPASSTHROUGH API. I'll copy the data to the Srb->DataBuffer.
		 */
		/* PayloadSize is the total Payload Size in DWORD. */
		if ( pStatus->PayloadSize*sizeof(U32) >= STATUS_PAYLOAD_HEAD_LENGTH )
		{
			VariableDataSize = (pStatus->PayloadSize*sizeof(U32))-STATUS_PAYLOAD_HEAD_LENGTH;
		}
		else
		{
			OSD_ASSERT( FALSE );
			VariableDataSize = 0;
		}
		OSD_DASSERT( VariableDataSize<=STATUS_PAYLOAD_VARIABLE_DATA_LENGTH );

		if ( VariableDataSize )	
		{
			OSD_TRACE(("Copy Payload Data: VariableDataSize=%d ScsiStatus=%d.\n", 
				VariableDataSize, Srb->ScsiStatus));
			OSD_DumpSrb(Srb);

			/* The Variable Data Field has data needs transfer. */
			if ( Srb->ScsiStatus==ssGOOD )
			{
				/* This is the PIO Data In Command. 
				 * Let me do the data transfer for firmware. */
				OSD_DASSERT( Srb->DataBuffer!=NULL );
				gen_memmove(
					Srb->DataBuffer,
					pStatus->VariableData,
					//VariableDataSize //TBD: Min(STATUS_PAYLOAD_VARIABLE_DATA_LENGTH, Srb->DataTransferLength)
					Min(VariableDataSize, Srb->DataTransferLength)
				);
			}
			else
			{
				/* Now it should be the Sensor Data. 
				 * Although firmware sends sense data but the SenseInfoBuffer can be NULL. */
				if ( Srb->SenseInfoBuffer!=NULL )
				{
					gen_memmove(
						Srb->SenseInfoBuffer, 
						pStatus->VariableData, 
						Min(STATUS_PAYLOAD_VARIABLE_DATA_LENGTH, Srb->SenseInfoBufferLength)
					);
				}
			}
		}

		/* Return to the OSD */
		pMUExtension->OutstandingRequest[pStatus->TaskTag] = NULL;
		Tag_Free(&pMUExtension->CmdTag, pStatus->TaskTag);
		OSD_DASSERT( pMUExtension->OutstandingRequestCount>=1 );
		pMUExtension->OutstandingRequestCount--;
		OSD_RAIDReturn(pOSDExtension, Srb);
	}

	MU_NotifyStatusTailChange(pMUExtension);

	return TRUE;
}

void MU_TriggerInterrupt( 
	IN bpMU_COMP_EXTENSION pMUExtension,
	IN U32 Interrupt 
)
{
	PTR_SIZE BaseAddress = (PTR_SIZE)pMUExtension->BaseAddress;
	OSD_HwWriteU32(
			BaseAddress + MU_INBOUND_DOORBELL,
			Interrupt);
}